--
-- SheetComb.hs -- Spreadsheet combinators
-- Owner -- Martin Erwig
--

module Etc.SheetComb where

import List ( (\\) )
import Maybe
-- import Data.Generics

import Etc.BiFunctor

-- import SpreadSheet


---------------------------------------------------------------------------
-- SPREADSHEETS COMBINATORS
---------------------------------------------------------------------------

-- 
-- Elementary combinators: one-cell spreadsheets (ocs)
--


-- unit lifts values to 1x1 spreadsheets
--
unit :: a -> Sheet a
unit x = Sheet [(origin,x)]

unitIV :: Int -> SheetV
unitIV = unit . I

unitIF :: Int -> SheetF
unitIF = unit . Inp . I

unitSV :: String -> SheetV
unitSV = unit . S

unitSF :: String -> SheetF
unitSF = unit . Inp . S


-- block is a 1x1 blocked spreadsheet
-- 
block :: SheetF
block = unit (Inp Unused)

-- a group of index-dependent spreadsheets:
-- 
-- posCell creates a one-cell spreadsheet that computes its value 
-- depending on its position (= indx).
--
-- Examples:
--   manhattan applied to an index creates a one-cell spreasheet that
--     stores the manhattan distance from the origin
--   tabulate applies a function to the values on the border of a 
--     given index
--     NOTE: The computation cannot be performed while constructing
--     a tabular cell. The delay can be achieved either by using functions
--     from the meta language (that is, creating a function on sheets) or
--     by creating a formula (of type Fml) that needs to be evaluated.
--
-- 
posCell :: (Indx -> a) -> [Indx] -> Sheet a
posCell f ixs = Sheet (map (\ix->(ix,f ix)) ixs)

-- posCell :: (Indx -> a) -> Indx -> Sheet a
-- posCell f ix = Sheet (ix,f ix)

manhattan :: [Indx] -> SheetV
manhattan = posCell (\ix->I (col ix+row ix-2))

-- tabulate :: (a -> a -> a) -> Indx -> Sheet a -> Sheet (Maybe a)
-- tabulate f ix s = posCell (\ix->... (s # ix) ...
-- 
-- tabulate :: BinOp -> [Indx] -> SheetF
-- tabulate b = posCell (\ix->Bin b (Ref (ix <! 1)) (Ref (ix >! 1)))
-- 
tabulate :: BinOp -> Indx -> Fml
tabulate b ix = Bin b (Ref (ix <! 0)) (Ref (ix >! 0))

mult :: Indx -> Fml
mult = tabulate Mult

-- table creates a spreadsheet by applying an index-dependent function
-- to all indices in the range of a given dimension
-- 
table :: (Indx -> a) -> Dim -> Sheet a
table f = posCell f . indxRange


--
-- Sheet transformers (structure/dimension preserving)
-- 
mapIV :: (Int -> Int) -> SheetV -> SheetV
mapIV f = fmap (\v->case v of {I i -> I (f i); e->e})

mapIF :: (Int -> Int) -> SheetF -> SheetF
mapIF f = fmap (\v->case v of {Inp (I i) -> Inp (I (f i)); f->f})


--
-- Joining and splitting spreadsheets horizontally and vertically
-- 
(<|>) :: (Eq a,Shift a) => Sheet a -> Sheet a -> Sheet a
s <|> s' | rowRng s == rowRng s' = s `cat` colShift (width s) s'
         | otherwise             = error "<|>: Row ranges don't agree"

(<->) :: (Eq a,Shift a) => Sheet a -> Sheet a -> Sheet a
s <-> s' | colRng s == colRng s' = s `cat` rowShift (height s) s'
         | otherwise             = error "<|>: Col ranges don't agree"

(>|<) :: Col -> Sheet a -> (Sheet a,Sheet a)
n >|< s = (clipR n s,clipL (n+1) s)

(>-<) :: Row -> Sheet a -> (Sheet a,Sheet a)
n >-< s = (clipT n s,clipB (n+1) s)


-- "save" joining: padd smaller sheet with blocked cells
-- 


{- LAWS

   diff (eval (s <|> s')) (eval s <|> eval s') 
   =
   errorCells (colShift (width s) s')

   assoc. laws:

   (s1 <|> s2) <|> s3  =  s1 <|> (s2 <|> s3)
   (s1 <-> s2) <-> s3  =  s1 <-> (s2 <-> s3)

   "spatial law" (distrib.):

   (s1 <-> s2) <|> (s3 <-> s4)  =  (s1 <|> s3) <-> (s2 <|> s4)
   iff <|> constraints hold for s1/s3 and s2/s4

   ...
  
   uncurry (<|>) (n >|< s) = s  
     only if neither of the split sheets does not contain an undefined
     row at the top or bottom

-}


--
-- Sequencing spreadsheets horizontally and vertically
--
--  f can be used to systematically create different versions of the input
--  sheet. f *must* be dimension preserving.
-- 
left :: (Eq a,Shift a) => (Sheet a -> Sheet a) -> Int -> Sheet a -> Sheet a
left f n s | n<1  = error "left: n must be at least 1"
           | n==1 = s
           | True = left f (n-1) (f s) <|> s 

right :: (Eq a,Shift a) => (Sheet a -> Sheet a) -> Int -> Sheet a -> Sheet a
right f n s | n<1  = error "right: n must be at least 1"
            | n==1 = s
            | True = s <|> right f (n-1) (f s)

up :: (Eq a,Shift a) => (Sheet a -> Sheet a) -> Int -> Sheet a -> Sheet a
up f n s | n<1  = error "up: n must be at least 1"
         | n==1 = s
         | True = up f (n-1) (f s) <-> s 

down :: (Eq a,Shift a) => (Sheet a -> Sheet a) -> Int -> Sheet a -> Sheet a
down f n s | n<1  = error "down: n must be at least 1"
           | n==1 = s
           | True = s <-> down f (n-1) (f s)


copyL :: (Eq a,Shift a) => Int -> Sheet a -> Sheet a
copyL = left id

copyR :: (Eq a,Shift a) => Int -> Sheet a -> Sheet a
copyR = right id

copyU :: (Eq a,Shift a) => Int -> Sheet a -> Sheet a
copyU = up id

copyD :: (Eq a,Shift a) => Int -> Sheet a -> Sheet a
copyD = down id



{- LAWS

   left n (left m s)  =  left (n*m) s
   down n (down m s)  =  down (n*m) s

   dim s = ((c,r),(c',r'))  ==>  dim (left n s) = ((c,r),(n*c',r'))
   dim s = ((c,r),(c',r'))  ==>  dim (down n s) = ((c,r),(c',n*r'))

-}


---------------------------------------------------------------------------
-- SPREADSHEETS ABSTRACTION
---------------------------------------------------------------------------

{-

Idea: spreadsheet abstraction is a mapping from a (detailed)
spreadsheet into a simpler one, omitting information.

Example: summarizing a spreadsheet calculation in a result
cell, represented as a one-cell spreadsheet. 

Further calculations can be expressed in terms of the abstractions
(and should be transformed into calculations of the underlying
detailed spreadsheet(s)).


-}

---------------------------------------------------------------------------
-- EXAMPLES
---------------------------------------------------------------------------

ab = Sheet [(a 1,i 3),(b 1,i 5)]

a2 = Sheet [(a 1,i 7),(a 2,i 11)]


natR n = right (mapIF succ) n (unitIF 1)
natD n = down (mapIF succ) n (unitIF 1)

binTable :: BinOp -> Col -> Row -> SheetF
binTable b c r = top <-> (border <|> body)
                     where top    = unitSF (show b) <|> natR c
                           border = natD r
                           body   = table (tabulate b) (ospan c r)

at99 = eval (binTable Plus 9 9)
mt99 = eval (binTable Mult 9 9)


