--
-- zhangUnits.hs -- zhangUnits
-- Owner -- ???
--

module Etc.DUT where 

import Util.PrintList  
{-

  value:    number, string, boolean, ..., error
            possible user-defined data types as in Forms/3

  content:  value, address, or expression


Cells contain "content"; cells are evaluated to "values".

-}

--
-- data model
-- 

---- generic sheets
---- 
type Address = String
type Cell a = (Address,a)

data Sheet a = Sheet [Cell a]

instance Functor Sheet where
  fmap f (Sheet cs) = Sheet (map (\(a,c)->(a,f c)) cs)

---- contents and values
---- 
data Value = Num Float
	   | Num1 Float
           | Str String
	   | Err String
	   deriving Eq

instance Show Value where
  show (Num i) = show i
  show (Num1 i) = show i
  show (Str s) = s
  show (Err s) = '!':s
   
data Content = Val Value
             | Ref Address
	     | Un  Op Content
	     | Bin Op Content Content
	     | Agg Op [Content] 
	     deriving Eq
	     
type Op = String

instance Show Content where
  show (Val v)     = show v
  show (Ref a)     = '@':a
  show (Un  f v)   = f++"("++show v++")"
  show (Bin f v w) = f++"("++show v++","++show w++")"
  show (Agg f vs)  = f++asList show vs
     
---- specific sheets (see also UnitSheet below)
---- 
type SpreadSheet = Sheet Content
type ValueSheet  = Sheet Value

lift :: ValueSheet -> SpreadSheet
lift = fmap Val

---- printing sheets
---- 
maxCol cs = maximum [a | ([a,_],_) <- cs]  
maxRow cs = maximum [i | ([_,i],_) <- cs]

colWidth c cs = maximum [length (show v) | ([a,_],v) <- cs, a==c]

padToWith i s c = s++take (2+i-length s) (repeat c)

padTo i s = padToWith i s ' '

pRow r cs = (if r=='1' then 
               "  | "++mkRow r cs (\c r cs->[c]) ' '++
	       "--+-"++mkRow r cs (\c r cs->"") '-'
             else "")++
               r:" | "++mkRow r cs pCell ' '

mkRow r cs f x = concat [padToWith (colWidth c cs) (f c r cs) x
	                 | c <- ['A'..maxCol cs]]++"\n"

pCell c r cs = maybe "" show (lookup [c,r] cs)

instance Show a => Show (Sheet a) where
    show (Sheet cs) = concat [pRow r cs | r <- ['1'..maxRow cs]]

---- pseudo constructors
---- 
n :: Float -> Content
n =  Val . Num

s :: String -> Content
s =  Val . Str

e :: String -> Content
e =  Val . Err

r :: String -> Content
r =  Ref

c :: Char -> Int -> Content -> Cell Content
c c i x = (c:show i,x)


add x y = Bin "+" (Ref x) (Ref y)
sub x y = Bin "-" (Ref x) (Ref y)  
mul x y = Bin "*" (Ref x) (Ref y)
divd x y = Bin "/" (Ref x) (Ref y)

subm xs = Agg "-" (map Ref xs)
divm xs = Agg "/" (map Ref xs)
mulm xs = Agg "*" (map Ref xs)
summ xs = Agg "+" (map Ref xs)
count xs = Agg "count" (map Ref xs)


row :: Int -> [Content] -> [Cell Content]
row i = zipWith (\c x->(c:show i, x)) ['A'..]


-- 
-- Sheet evaluator
-- 
eval :: SpreadSheet -> ValueSheet
eval (Sheet cs) = Sheet (map (\(a,c)->(a,evalC cs c)) cs)

evalC :: [Cell Content] -> Content -> Value
evalC cs (Val v) = v
evalC cs (Ref a) = maybe (Err a) (evalC cs) (lookup a cs)
evalC cs (Agg f xs)  = evalAgg f (map (evalC cs) xs)

evalAgg f xs = case typeOf f of
                  "Num" -> maybe (Err f) Num (evalAggNum f xs)
		  _     -> Err f

typeOf "+" = "Num"
typeOf "-" = "Num"	
typeOf "*" = "Num"
typeOf "/" = "Num"
typeOf "count" = "Count" 
typeOf _   = "Unknown"

binFun "+" = Just (+)
binFun "-" = Just (-)  
binFun "*" = Just (*) 
binFun "/" = Just (/)
binFun _   = Nothing

signOf "/" = "/-"
signOf "-" = "/-"

evalAggNum :: String -> [Value] -> Maybe Float
evalAggNum f xs = do ys <- expectNums xs
                     g <- binFun f
		     return (foldl1 g ys)
evalAggNum _ _  = Nothing

expectNums :: [Value] -> Maybe [Float]
expectNums vs = mapM (\v->case v of {Num x->Just x; _->Nothing}) vs


--
-- Types and Units
-- 
data Type = NumT | StrT | TypeError
            deriving (Eq,Show)

data Unit = Base      String
          | Dependent Unit Unit    -- bottom-up 
	  | And       [Unit]
	  | Or        [Unit]
	  | One
	  
--instance Fractional Int where
--  (/) :: a ->a ->a 
 
 
instance Eq Unit where
  Base u        == Base v        = u == v
  Dependent u v == Dependent w x = u == w && v == x
  And us        == And vs        = all (`elem` vs) us && all (`elem` us) vs
  Or us         == Or vs         = all (`elem` vs) us && all (`elem` us) vs
  One          == One          = True
  _             == _             = False

---- normalization of unit expressions
---- 
normU (Or [Dependent u' u,Dependent v' v]) | u==v = 
          Dependent (normU (Or [u',v'])) u
          
normU (Or ((And [Dependent u' u, Dependent v' v]):
		(And [Dependent w' w,Dependent x' x]):andU))
    |u==w||v==x = normU (Or ((And [normU (Or [Dependent u' u,Dependent w' w]),
                     normU (Or [Dependent v' v,Dependent x' x])]):andU))
    |u==x||v==w = normU (Or ((And [normU (Or [Dependent u' u,Dependent x' x]),
                     normU (Or [Dependent v' v,Dependent w' w])]):andU))
                     
normU (Or [u,One]) = u
normU (Or [One,u]) = u
normU (Or [u,v]) | u==v = u
normU (Or [(Or (c:cs)),v])| cs == [] = normU (Or [c,v]) 
		          | c==v = Or (c:cs)
		          | otherwise= Or [c, normU (Or [(Or cs),v])]
		          
normU (And [Dependent u' u, Dependent v' v])| u==v && u'==v' = Dependent u' u		          
normU (And [One,u]) = u
normU (And [u,One]) = u		          
normU u = u

-- | ***********************************************************************
-- This doesn't work
simpfy :: UnitSheet -> Headers -> SpreadSheet ->UnitSheet
simpfy (Sheet us) hs (Sheet cs) = Sheet (map (\(a,c)->(a,equit c hs cs)) us)
 
equit :: Unit -> Headers -> [Cell Content] ->Unit
equit (And [Dependent us u, Dependent vs c]) hs  cs
       |us == getAllU u hs cs = And [u, Dependent vs c] 
	-- \|us == getAllU u hs cs && vs == getAllU c hs cs = And [u,c]
	-- \|us == getAllU u hs cs = And [u, Dependent vs c]
       |vs == getAllU c hs cs = And [Dependent us u, c]
      |otherwise = And [Dependent us u, Dependent vs c]
equit (Dependent us u) hs  cs = Dependent us u
equit (Base u) hs  cs = Base u
equit one hs  cs = One

getAllU :: Unit ->Headers -> [Cell Content] -> Unit
getAllU u hs cs = normU (Or [(getSU a cs)|(a,[u'])<- hs, str==u'])
			where str = lookupAddr (show u) cs

getSU a cs = case lookup a cs of
		Just z -> Base (show z)
		_	-> One

lookupAddr :: String ->[Cell Content]->Address
lookupAddr u xys = head [x|(x,y) <- xys, Val(Str u) == y]
-- | ***********************************************************************

instance Show Unit where
  show (Base s)         = s
  show (Dependent s u)  = show u++"["++show s++"]"
  show (And us)         = printList ["","&",""] show us
  show (Or us)          = printList ["","|",""] show us
  show One             = "_"

type TypeA = [(Address,Type)]

type Headers = [(Address,[Address])]

---- Unit inference
----
lookupStr :: Address -> [Cell Content] -> Maybe String
lookupStr a cs = case lookup a cs of 
                    Just (Val (Str s)) -> Just s
		    _                  -> Nothing

type UnitSheet = Sheet Unit

inferU :: Headers -> SpreadSheet -> UnitSheet
inferU hs (Sheet cs) = Sheet (map (\(a,c)->(a, inferUC cs hs (a,c))) cs)

--inferU hs (Sheet cs) = Sheet (map (\(a,c)->(a, (equit (inferUC cs hs (a,c))
--  hs cs)))  cs)                                                                     

inferUC :: [Cell Content] -> Headers -> Cell Content -> Unit
inferUC cs hs (a,Val v) = getUnit cs hs a
-- first check the header info
-- no header then check Ref
inferUC cs hs (a,Ref b) = case lookup a hs of		   
			    Just v -> getUnit cs hs a
			    _ 	   -> case lookup b cs of  
			    		Just k -> inferUC cs hs (b,k)
			    		_      -> One

inferUC cs hs (a,Agg "+" xs)  = normU (And [bu,hu])
			    where hu = getUnit cs hs a
			    	  bu = normU (Or (map (inferUV cs hs) xs))
			    	  
inferUC cs hs (a,Agg "*" (x:xs))
      |xs == [] = hunit
      |unit1 /= One && all(== One)(map (inferUV cs hs)xs)
	      = And [unit1, hunit]
      |unit1 == One = inferUC cs hs (a,Agg "*" xs)
      |otherwise = hunit
				  where unit1 = inferUV cs hs x
					hunit = getUnit cs hs a  
				   				   
inferUC cs hs (a,Agg  f (x:xs))
      = case signOf f of
	 "/-" -> case x of
	          Ref b ->normU (And [(inferUC cs hs (a, Ref b)),(getUnit cs hs a)])
								  	     				
		  _     -> One			   		
	 _      -> One  			

inferUC _ _ _ = One


inferUV :: [Cell Content] -> Headers -> Content -> Unit
inferUV cs hs (Val v) = One
inferUV cs hs (Ref a) = case lookup a cs of
                            Just v -> inferUC cs hs (a,v)
			    _      -> One
inferUV _ _ _ = One

getUnit :: [Cell Content] -> Headers -> Address -> Unit
getUnit cs hs a = case lookup a hs of
		     Nothing  -> One
		     Just []  -> One
-- 		     Just [b] -> maybe One Base (lookupStr b cs)
		     Just bs -> traceUnits cs hs (zip bs (map (getU cs) bs))

getU cs b = maybe One Base (lookupStr b cs)

traceUnits :: [Cell Content] -> Headers -> [(Address,Unit)] -> Unit
traceUnits cs hs bus = case map (traceUnit cs hs) bus of
                         [u] -> u
			 us  -> And us

traceUnit :: [Cell Content] -> Headers -> (Address,Unit) -> Unit
traceUnit cs hs (a,u) = case getUnit cs hs a of
                          One -> u
			  v    -> Dependent u v    

--
-- Example
--

{-
    A      B      C     D       E
1        Fruit
2  Month Apple   Orange Plum    Total 
3    May     8     11    21     *  
4   June    75     5     15     /
5   July    30     @B3   30     +
6    @D2   SUM    SUM  SUM      @Z9
7          @C3    

-}

harvest :: SpreadSheet
harvest = Sheet (
  c 'B' 1 (s "Fruit"):
  row 2 (map s ["Month","Apple","Orange","Plum","Total"])++
  row 3 [s "May", n  8,n 11, n 21, mulm ["B3", "C3","D3"]]++
  row 4 [s "June", n 75,n  5, n 15, divm ["B4", "C4","D4"]]++
  row 5 [s "July",n 30,r "B3", n 30, summ ["B5","C5","D5"]]++
  row 6 [r "D2",summ ["B3","B4","B5"],
                summ ["C3","C4","C5"],
		summ ["D3","D4","D5"],r "Z9"]++
  [c 'B' 7 (r "C3")])

-- for testing ...
-- 
Sheet h = harvest 

result :: ValueSheet
result = eval harvest

-- not needed yet ...
-- harvestTypeA = []

-- information about headers gathered from the user
-- 
harvestHeaders = [("B1",[]),("B2",["B1"]),("C2",["B1"]),("D2",["B1"]),
                  ("A2",[]),("A3",["A2"]),("A4",["A2"]),("A5",["A2"]),
                  ("B3",["A3","B2"]),("C3",["A3","C2"]),("D3",["A3","D2"]),
                  ("B4",["A4","B2"]),("C4",["A4","C2"]),("D4",["A4","D2"]),
                  ("B5",["A5","B2"]),("C5",["A5","C2"]),("D5",["A5","D2"])] 
units :: UnitSheet
units = inferU harvestHeaders harvest
sunits = simpfy units harvestHeaders harvest

-- we only have a form of weak subject reduction:
-- 
unitsR :: UnitSheet
unitsR = inferU harvestHeaders (lift result)

