--
-- ExcelParser.hs -- Excel formula parser
-- Owner -- Robin Abraham
--

-- Excel formula parser using monadic parser combinators

module Sheet.ExcelParser where

import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Expr
import Text.ParserCombinators.Parsec.Token
--import Text.ParserCombinators.Parsec.Perm
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language 
import List (intersperse)
import Char (digitToInt,ord,isAlpha,isDigit)
import Sheet.Expr
import Sheet.Indx

-----------------------------------------------------------
-- functions to run stuff
-----------------------------------------------------------
run :: Show a => Parser a -> String -> IO ()
run p input
        = case (parse p "" input) of
            Left err -> do{ putStr "parse error at "
                          ; print err
                          }
            Right x  -> print x

-- runfml :: Show a => Parser a -> String -> 
runfml p input
        = case (parse p "" input) of
            Left err -> UnkFml
            Right x  -> x


-------------------------------------------------------------
-- Formula grammar being used in current version
-------------------------------------------------------------
{-
formula ::= "="expression
expression ::= expression logop expr1 | expr1
expr1 ::= expr2 & expr1 | expr2
expr2 ::= expr2 aop expr3 | expr3
expr3 ::= expr3 mop expr4 | expr4
expr4 ::= expr4 eop expr5 | expr5
expr5 ::= expr5"%" | expr6
expr6 ::= "-"expr6 | "+"expr6 | expr7
expr7 ::= reference | function | constant | "(" expression ")"
          | namedref | cellref
function ::= fname "(" arguments ")"
arguments ::= expression "," arguments | expression | epsilon
reference ::= address | address ":" address | name
cellref ::= rangeref | func ":" func | rangeref ":" func | func ":" rangeref
rangeref ::= sheetref address 
             | workbookref stringorid ":" stringorid "!" address
             | rangeref ":" rangeref | colrowref | address
namedref ::= sheetref String | workbookref String
stringorid ::= String | Id
logop ::= "=" | "<" | "<=" | ">" | ">=" | "<>"
aop ::= "+" | "-"
mop ::= "*" | "/"
eop ::= "^"

-}


formula = do { char '=';
               e <- expression;
               return (e);  
	       }
      <|> do { e <- constant;
	       return e;
	     }
          <?> "formula level"

expression = do { -- e1 <-  expression;
                  -- l <- logop;
                  e2 <- try(expr1 `chainl1` logop);
		  return e2; -- (Log l e1 e2);
                }
	 <|> do { e1 <- expr1;
                  return e1;  
                }
         <?> "expression"

expr1 = {-do { e1 <- expr2;
             char '&';
	     e2 <- expr1;
             return expr1;  -- modify later
           }
    <|> -} do { e1 <- expr2;
             return e1; 
           }
    <?> "expr1"

expr2 = do { -- e1 <- expr2;
             -- a  <- aop;
             e2 <- try (expr3 `chainl1` aop); -- expr3;
             return e2; -- (Bin a e1 e2); 
           }
    <|> do { e1 <- expr3;
             return e1;
           }
    <?> "expr2"

expr3 = do { -- e1 <- expr3;
             -- m  <- mop;
             e2 <- try (expr4 `chainl1` mop); -- expr4;
             return e2; -- (Bin m e1 e2);    
           }
    <|> do { e1 <- expr4;
             return e1;   
           }
    <?> "expr3"

expr4 = do { -- e1 <- expr4;
             -- e  <- eop;
             e2 <- try (expr5 `chainl1` eop); -- expr5;
             return e2; -- (Bin e e1 e2);   
           }
    <|> do { e1 <- expr5;
             return e1;   
           }
    <?> "expr4"
 
expr5 = {- do { e1 <- expr5;
             char '%';
             return e1;  -- modify later
           }
    <|> -} do { e1 <- expr6;
             return e1; 
           }
    <?> "expr5"

expr6 = do { char '-';      
             e1 <- expr6;
             return (Bin Minus (Inp (I 0)) e1);
           }
    <|> do { char '+';
             e1 <- expr6;
             return (Bin Plus (Inp (I 0)) e1); 
           }
    <|> do { e1 <- expr7;
             return e1;  
           }
    <?> "expr6"

expr7 = do { r <- reference;
             return r; 
           }
   <|>  do { f <- function;
             return f; 
           }
   <|>  do { c <- constant;
             return c; 
           }
   <|>  do { char '(';
             e <- expression;
             char ')';
             return e; 
           }
{--   <|>  do { n <- namedref;
             return n; 
           } 
   <|>  do { c <- cellref;
             return c; 
           } --}
   <?>  "parse error in expr7"

function = do { f <- try (handledFunctions);
                char '(';
                a <- arguments;
                char ')';
                return (f a);
              }
        <|> do { try (string "IF(");
		 l <- expression;
		 char ',';
		 e1 <- expression;
                 char ',';
                 e2 <- expression;
		 char ')';    
	         return (If l e1 e2);
               }
        <|> do { f <- try(word);        -- for general case
                 char '(';
		 a <- arguments;
                 char ')';
                 return (Func f a);
               }
        <?> "function"

handledFunctions = do { f <- try (string "SUM");
			return (Agg Plus);
                      }
              <|>  do { f <- try (string "AVERAGE");
			return (Agg Avg);
		      }
	      <|>  do { f <- try (string "AND");
			return (AggL And);
		      }
	      <|>  do { f <- try (string "OR");
			return (AggL Or);
		      }
 {-             <|>  do { f <- try (string "IF");
			return If;
		      } -}
	      <?> "handled functions"

arguments = do { e <- try (commaExpression);
		 return e;
               }
	<|> do { e <- try (expression);
                 return [e];
	       }
--	<|> do { return [];   -- case for epsilon
--		}
        <?> "arguments"

commaExpression = do { e <- expression;
                       char ',';
		       as <- arguments;
		       return (e:as);
		     }

reference = do { a <- try (colRange);
                 return (RefRng a); 
               }
       <|>  do { a <- try(address);
                 return (Ref a);
	       }
{-       <|>  do { n <- word;   -- for named reference
                 return n;   -- modify later
               } -}
       <?> "reference"

colRange =  do { a1 <- address;
                 char ':';
		 a2 <- address;
		 return (genRefs a1 a2);
	       }

{-- Include in parser after expanding data type
cellref = do { r <- rangeref;
               return r;   -- modify later
             }
      <|> do { f1 <- function;   -- func
               char ':';
               f2 <- function;   -- func
               return f1;   -- modify later
             }
      <|> do { r <- rangeref;
               char ':';
               f <- function;   -- func
               return r;   -- modify later
             }
      <|> do { f <- function;   -- func
               char ':';
               r <- rangeref;
               return f;   -- modify later
             }
      <?> "cellref"
--}

{-- Include this after expanding datatype
rangeref = do { s <- sheetref;
		a <- address;
		return a;   -- modify later
	      }
       <|> do { w <- workbookref;
		s1 <- stringorid;
		char ':';
		s2 <- stringorid;
		char '!';
		a <- address;
		return w;  -- modify later
	      }
       <|> do { r1 <- rangeref;
		char ':';
		r2 <- rangeref;
		return r1;  -- rangeref should return a range
	      }
       <|> do { c <- address; -- this should be colrowref
		return c;   -- colrowref should return a reference
	      }
       <|> do { a <- address;
		return a;  -- address should return a reference already
	      }
       <?> "rangeref"
--}

namedref = do { s <- sheetref;
		w <- word;
		return s; -- modify later
	      }
       <|> do { w <- workbookref;
		s <- word;
		return w; -- modify later
	      }
       <?> "namedref"

stringorid = do { s <- word;
		  return s;
		}
         <|> do { i <- word;
		  return i;
		}
	 <?> "string or id"

workbookref = do { s <- word;
		   return s;
		 }
          <?> "Workbook reference"

sheetref = do { s <- word;
		   return s;
	      }
       <?> "Worksheet reference"

address = do { s <- try(count 2 upper); -- for case beyond col Z
               i <- try(many1 digit);
               return (adConv (s++i));
	     }
      <|> do { s <- try(upper);  
               i <- try(many1 digit);  
               return (adConv (s:i));
             }
     <?>  "address parsing"

constant = do { char '"';
		s <- manyTill anyToken (char '"');
		return (Inp (S s));
              }
       <|> do { f <- try(Sheet.ExcelParser.float);
                return (Inp (F f));
	      }
       <|> do { i <- try(Sheet.ExcelParser.integer);
                return (Inp (I (fromEnum i)));
	      }
{-       <|> do { f <- try (many1 digit);
		try (char '.');
		g <- try (many1 digit);
		return (Inp (F (read (f++"."++g))));
	      }
       <|> do { i <- many1 digit; 
                return (Inp (I (read i)));
	      } -}
       <?> "constant parser"


word = many1 letter <?> "word parser"


lexer = P.makeTokenParser haskellDef
float = P.float lexer
integer = P.integer lexer
comma = P.comma lexer;
colon = P.colon lexer;
wSp = P.whiteSpace lexer;
--alphanum = P.Char.alphanum lexer;
str = P.stringLiteral lexer;


logop = do { char '=';
	   return (Log Eq); 
	 }
  <|> do { char '<';
	   return (Log Lt);  
	 }
  <|> do { char '<';
	   char '=';
	   return (Log Lt);  -- modify later to Le
	 }
  <|> do { char '>';
	   return (Log Gt); 
	 }
  <|> do { char '>';
	   char '=';
	   return (Log Gt);  -- modify later Ge
	 }
  <|> do { char '<';
	   char '>';
	   return (Log Neq);  
	 }
  <?> "logical operation"


aop = do { char '+';
	   return (Bin Plus);  
         }
  <|> do { char '-';
	   return (Bin Minus); 
	 }
  <?> "plus or minus"

mop = do { char '*';
	   return (Bin Mult); 
	 }
  <|> do { char '/';
	   return (Bin Div);  
	 }
  <?> "prod or div"

eop = do { char '^';
	   return (Bin Exp); 
	 }
    <?> "exponentiation"
	   

------------------------
-- auxilliary functions
------------------------

-- genRefs builds the list of cells references by a range 
genRefs :: Indx -> Indx -> [Fml]
genRefs (Indx c1 r1) (Indx c2 r2) = map Ref [Indx c r|c <- [c1..c2], r <- [r1..r2]]

genRefs' :: String -> [Fml]
genRefs' s = genRefs (adConv i1) (adConv i2)
             where ss = lex s
		   i1 = fst (head ss)
		   i2 = tail (snd (head ss))

-- colConv converts the column value from Excel format to the backend
-- format. ie, From A to 1
colConv :: String -> Int
colConv s = foldr (+) 0 (zipWith (*) p (map (\x -> (ord x -64)) s'))
	    where s' = reverse s
		  p = map (26^) [0..(length s)-1]

-- adConv converts address values not seperated by spaces.
adConv :: String -> Indx
adConv s = Indx (colConv colVal) (foldr (+) 0 p')
	   where colVal = getCol s
		 rowVal = reverse (map digitToInt (getRow s))
		 p = map (10^) [0..(length rowVal)-1]
		 p' = zipWith (*) p rowVal

getCol :: String -> String
getCol [] = []
getCol (s:ss) | isAlpha s = s:(getCol ss)
	      | isDigit s = []

getRow :: String -> String
getRow [] = []
getRow (s:ss) | isDigit s = s:(getRow ss)
              | isAlpha s = getRow ss
