-------------------------------------------------------------------
-- Expr has the data types for representing the Excel formulas in 
-- the backend inference engine
-- Written by: Robin Abraham
-- Added to CVS on 9/26/2005
-------------------------------------------------------------------

module Sheet.Expr where

import Sheet.Indx
--import Sheet.Sheet
import List (intersperse,nub)
import Maybe (fromMaybe)
import Debug.Trace

---------------------------------------------------------------------------
-- CELL CONTENTS
---------------------------------------------------------------------------

data Val = I Int
         | F Double
         | S String
         | B Bool
         | E String  -- Error / undefined
         | Unused    -- block out `dead parts'
           deriving Read

valNegate :: Val -> Val
valNegate (I i) = I (-i)
valNegate (F f) = F (-f)
valNegate (B b) = B (not b)
valNegate v = v

data Fml = Inp Val            -- input (by user)
         | Ref Indx           -- direct reference to another cell
         | RefRng Rng         -- this is hack to help parser *** remove this later
	 | Outer Fml          -- this is dummy to pretty-print the "=" for Excel
         | Bin BinOp Fml Fml  -- application of binary operation
         | Agg BinOp Rng
	 | AggL LogOp Rng
         | If Fml Fml Fml
         | Log LogOp Fml Fml
         | Func String [Fml]   -- general case of formula with name and args
         | UnkFml
           deriving (Eq,Ord)

data BinOp = Plus | Minus | Mult | Div | Avg | Exp
             deriving (Eq,Ord,Read)

data LogOp = Eq | Neq | Lt | Gt | And | Or
             deriving (Eq,Ord,Read)


-- instance Data Int where

instance Show Val where
  show (I i)  = show i
  -- show (S s)  = "\""++s++"\""
  -- a comentada era a original...
  show (S s)  = s
  show (B b)  = show b
  show (F d)  = show d
  show (E e)  = e
  show Unused = "xxx"


instance Eq Val where
  (I x) == (I y) = x == y 
  (F x) == (F y) = abs (x - y) < 0.1
  (F x) == (I y) = abs (x - toEnum y) < 0.1
  (I x) == (F y) = abs (toEnum x - y) < 0.1		   
  (S x) == (S y) = x == y
  (B x) == (B y) = x == y
  _     == _     = False  



instance Ord Val where
  (I x) <= (I y) = x <= y 
  (F x) <= (F y) = x <= y
  (I x) <= (F y) = (toEnum x) <= y
  (F x) <= (I y) = x <= (toEnum y)
  _     <= _     = False  


{- [scott] 4-13 commented after change martin made to Indx.hs
instance Shift Val where {}
-}

instance Show Fml where
  show (Inp v)     = show v
  show (Ref ix)    = show ix
  show (Outer f)   = '=':show f
  show (Bin b f g) = "("++show f++" "++show b++" "++show g++")"
  show (Agg b r)   = --  trace "Agg" 
											fshow b++"("++rshow r++")"
  show (AggL l r)   = -- trace "AggL" 
											show l++"("++rshow r++")"
  show (If b e1 e2) = "IF("++show b++","++show e1++","++show e2++")"
  show (Log l e1 e2) = show e1++show l++show e2 
  show (Func s rs) = -- trace "Func" 
													s++"("++rshow rs++")"
  show (UnkFml) = "XXX"
  show (RefRng rs) = -- trace "RefRng"
													concatMap show rs

{- [scott] 4-13 commented after change martin made to Indx.hs
instance Shift Fml where
  shift dx (Ref ix)    = Ref (shift dx ix)
  shift dx (Bin b f g) = Bin b (shift dx f) (shift dx g)
  shift _  f           = f
-}

instance Show BinOp where
  show Plus  = "+"
  show Minus = "-"
  show Mult  = "*"
  show Div   = "/"
  show Exp   = "^"
  show Avg   = "avg"
--  show SlashM = "/-"
--  show Slash = "/"


fshow :: BinOp -> String
fshow Plus = "SUM"
fshow Mult = "PROD"
fshow Avg = "AVERAGE"

rshow :: [Fml] -> String
rshow [] = ""
rshow xs = foldr1 (++) (intersperse "," (map show xs))

instance Show LogOp where
  show Eq = "="
  show Neq = "<>"
  show Lt = "<"
  show Gt = ">"
  show And = "AND"
  show Or = "OR"


type Rng = [Fml] 

lop :: (Ord a) => LogOp -> (a -> a -> Bool)
lop Eq = (==)
lop Neq = (/=)
lop Lt = (<)
lop Gt = (>)

