--
-- Sheet.hs -- Generic sheets
--

module Sheet.Sheet where

import List ( (\\), union, intersect, nub)
import Maybe

import Sheet.Indx


---------------------------------------------------------------------------
-- SPREADSHEET TYPES
---------------------------------------------------------------------------

--
-- Type definitions
-- 
type Cell a = (Indx,a)

newtype Sheet a = Sheet {unSheet::[Cell a]}
                   deriving Eq


-- 
-- Auxiliary functions
-- 
( # ) :: Sheet a -> Indx -> Maybe a
Sheet cs # ix = lookup ix cs

maxIndx :: Sheet a -> Indx
maxIndx s = Indx (maximum (map col xs)) (maximum (map row xs))
            where xs = dom s

minIndx :: Sheet a -> Indx
minIndx s = Indx (minimum (map col xs)) (minimum (map row xs))
            where xs = dom s

-- 
-- Functor and Show instances
-- 
instance Functor Sheet where
  fmap f (Sheet cs) = Sheet (map (\(i,c)->(i,f c)) cs)

instance Show a => Show (Sheet a) where
  show (Sheet []) = "<empty>"
  show s          = showSheet s

showSheet :: Show a => Sheet a -> String
showSheet s = showHdr w (minCol,maxCol) ++
              unlines (map (showRow s w (minCol,maxCol)) [minRow..maxRow])
               where Indx maxCol maxRow = maxIndx s
                     Indx minCol minRow = minIndx s
                     w = maximum (map (length . show) (rng s))

showHdr :: Int -> ColRng -> String
showHdr w (c,c') = "    "++concatMap (\c->" | "++center w (shwCol c)) cols++"\n"++
                   replicate (4+length cols*(3+w)) '-'++"\n"
                     where cols = [c..c']

showRow :: Show a => Sheet a -> Int -> ColRng -> Row -> String
showRow s w (c,c') r = p++concatMap (showCell w) cs
                       where p = padLeftTo 4 (shwRow r)
                             cs = map ((s # ) . (flip Indx r)) [c..c']

showCell :: Show a => Int -> Maybe a -> String
showCell w mc = " | "++center w (maybe "" show mc)

center :: Int -> String -> String
center k s = spaces i++s++spaces j
             where n = length s
                   i = (k-n) `div` 2
                   j = k-n-i

padLeftTo :: Int -> String -> String
padLeftTo k s = spaces (k-length s)++s

spaces :: Int -> String
spaces n = replicate n ' '


---------------------------------------------------------------------------
-- BASIC, GENERIC SPREADSHEET OPERATIONS
---------------------------------------------------------------------------


dom :: Sheet a -> [Indx]
dom (Sheet cs) = map fst cs

rng :: Sheet a -> [a]
rng (Sheet cs) = map snd cs


-- Comparing spreadsheets
--
diff :: Eq a => Sheet a -> Sheet a -> [Indx]
diff (Sheet cs) (Sheet cs') = [fst c | c <- cs, notElem c cs'] `union`
                              [fst c | c <- cs', notElem c cs] 

-- concatenate two non-overlapping spreadsheets
-- 
cat :: Eq a => Sheet a -> Sheet a -> Sheet a
cat s@(Sheet cs) s'@(Sheet cs') | dom s `intersect` dom s'==[] = Sheet (cs++cs')
                                | otherwise = error ("cat: Common cells")

-- merge two possibly overlaping spreadsheets
-- 
-- merge :: Eq a => (a -> a -> a) -> Sheet a -> Sheet a -> Sheet a
-- merge f (Sheet cs) (Sheet cs') = ...


-- rngFilter removes all cells whose value does not satisfy the given predicate
--
rngFilter :: (a -> Bool) -> Sheet a -> Sheet a
rngFilter p (Sheet cs) = Sheet [c | c <- cs, p (snd c)]

-- normalize shifts a spreadsheet to start at the origin
--
-- normalize :: Shift a => Sheet a -> Sheet a
-- normalize s = shift (1-c,1-r) s
--              where  Indx c r = minIndx s

