--
-- HdrAux.hs -- Auxilliary functions for header inference
-- Owner -- Scott Andrew
--

module UCheck.HdrAux where

import Sheet.Indx
import Sheet.Sheet
import Sheet.Expr
import List
import UCheck.HdrDef
import Maybe

-- Functions for checking the type of values in cells.
-- isString checks for strings
isString :: Fml -> Bool
isString (Inp (S _)) = True
isString _ = False

isBlank :: Fml -> Bool
isBlank (Inp (S " ")) = True
isBlank _ = False

-- isNum checks for numbers
isNum :: Fml -> Bool
isNum (Inp (I _)) = True
isNum (Inp (F _)) = True
isNum _ = False


-- isAgg checks for aggregations
isAgg :: Fml -> Bool
isAgg (Agg _ _) = True
isAgg _ = False

-- isAgg' checks Bin for references and counts them as aggregations(lower conf.)
isAgg' :: Fml -> Bool
isAgg' (Bin _ f1 f2) = (isAgg' f1) || (isAgg' f2)
isAgg' (Ref _) = True
isAgg' _ = False

-- isBin checks for Bin (very low confidence for footer)
isBin :: Fml -> Bool
isBin (Bin _ _ _) = True
isBin _ = False

-- isRef checks for references
isRef :: Fml -> Bool
isRef (Ref _) = True
isRef _ = False

-- isUnused checks for cells that are unused.
isUnused :: Fml -> Bool 
isUnused (Inp Unused) = True
isUnused _ = False

-- isUnk checks for cells that are unparsed.
isUnk :: Fml -> Bool
isUnk f = f == UnkFml

-- aggEx extracts the indices of the cells participating in aggregation
aggEx :: Fml -> [Indx] 
aggEx (Agg _ rs) = concatMap aggEx rs
aggEx (Ref i) = [i] 
aggEx (Bin _ f1 f2) = aggEx f1 ++ aggEx f2
aggEx _ = [] 

-- neighbors generates list of cells in immediate neighborhood of given cell
-- this assumes the spreadsheet starts at (Indx 1 1) - no negatives for row/col
neighbor :: Indx -> [Indx]
neighbor (Indx c r) = filter (\x -> col x >= 1 && row x >= 1) ns \\ [Indx c r]
                      where ns = [Indx x y| x <- [c-1..c+1], y <- [r-1..r+1]]

topNeighbor :: Indx -> [Indx]
topNeighbor (Indx c r) = [Indx x (r-1)|x <- [c-1..c+1]]

botNeighbor :: Indx -> [Indx]
botNeighbor (Indx c r) = [Indx x (r+1)|x <- [c-1..c+1]]

leftNeighbor :: Indx -> [Indx]
leftNeighbor (Indx c r) = [Indx (c-1) y|y <- [r-1..r+1]]

rightNeighbor :: Indx -> [Indx]
rightNeighbor (Indx c r) = [Indx (c+1) y|y <- [r-1..r+1]]


-- compFml compares formulae for type equality
-- ie, it does not evaluate to values.
comp :: Sheet Fml ->  Maybe Fml -> Maybe Fml -> Bool
comp s f1 f2 = compFml f1' f2'
               where f1' = checkRef s (maybe (Inp Unused) id f1)
                     f2' = checkRef s (maybe (Inp Unused) id f2)
			

compFml :: Fml -> Fml -> Bool
compFml (Inp x) (Inp y) = compInp x y
compFml (Bin _ _ _) (Bin _ _ _) = True
compFml (Agg _ _) (Agg _ _) = True
compFml _ _ = False

compInp :: Val -> Val -> Bool
compInp (I _) (I _) = True
compInp (F _) (F _) = True
compInp (I _) (F _) = True
compInp (F _) (I _) = True
compInp (S _) (S _) = True
compInp (E _) (E _) = True
compInp Unused Unused = True
compInp _ _   = False

-- checkRef finds the cell the current cell is referencing
checkRef :: Sheet Fml -> Fml -> Fml
checkRef s (Ref x) = checkRef s (maybe (Inp Unused) id (s # x))
checkRef s x = x

-- Auxilliary functions for fences
--MM isNeighbor finds all cells one-cell distance away (even diagonally) as 
-- neighbors. In general, we only conside top, bottom, left and right
-- neighbors - is more accurate.
isNeighbor :: Indx -> Indx -> Bool
isNeighbor ix iy = dist ix iy <= 1

-- dist can be modified to exclude diagonal cells as neighbors by
-- changing the max to a sum of the two distances.
dist :: Indx -> Indx -> Int
dist ix iy = max (abs (col ix - col iy)) (abs (row ix - row iy)) 

-- softDist does not consider diagonal neighbors.
-- Positive result means first cell is to the left or above the 
-- second cell and vice versa for negative result.
softDist :: Indx -> Indx -> Int
softDist ix iy = dh + dv
		 where dh = col iy - col ix
		       dv = row iy - row ix

-- isSoftNeigbor checks to see if the second cell is across a soft fence from 
-- the first cell - we're moving from left to right and top to bottom
isSoftNeighbor :: Sheet Fml -> Indx -> Indx -> Bool
isSoftNeighbor s ix iy = isAgg fx && (isString fy || isNum fy) && (softDist ix iy == 1)
                                     where fx = fromMaybe UnkFml (s # ix) 
                                           fy = fromMaybe UnkFml (s # iy)

-- End of auxilliary functions for fences

-- "Distance" and "Cost" functions.
compareDist :: (Int,Indx) -> (Int,Indx) -> Ordering
compareDist (d1,_) (d2,_) = compare d1 d2

rowDist :: Indx -> Indx -> Int
rowDist (Indx _ r1) (Indx _ r2) = (r1-r2)

colDist :: Indx -> Indx -> Int
colDist (Indx c1 _) (Indx c2 _) = (c1-c2) 

type Cost = Int

rowCost :: Indx -> [Indx] -> Cost
rowCost x xs = sum (map (\y-> abs (colDist x y) - 5) xs)

colCost :: Indx -> [Indx] -> Cost
colCost x xs = sum (map (\y-> abs (rowDist x y) - 5) xs)


byRow :: [Indx] -> [(Row,[Indx])]
byRow [] = []
byRow (x:xs) = (row x,x:(xs')):(byRow xs'')
               where xs' = filter (\y -> row y == row x) xs
                     xs'' = xs \\ xs'

byCol :: [Indx] -> [(Col,[Indx])]
byCol [] = []
byCol (x:xs) = (col x,x:(xs')):(byCol xs'')
               where xs' = filter (\y -> col y == col x) xs
                     xs'' = xs \\ xs'

-- rowRng gives the min and max row numbers of a list of Indx
rowRng :: [Indx] -> (Int,Int)
rowRng [] = (0,0)
rowRng is = (minimum rs, maximum rs)
           where rs = map (\x -> row x) is

-- colRng gives the min and max column numbers of a list of Indx
colRng :: [Indx] -> (Int,Int)
colRng [] = (0,0)
colRng is = (minimum rs, maximum rs)
           where rs = map (\x -> col x) is

--MM flipHeader reverses the direction of is-header-of to is-child-of
-- or vice versa. It is an auxilliary function that is pretty handy if
-- we need to change our view.
flipHeader :: Headers -> Headers
flipHeader = regrpHdr.flipper

flipper :: Headers -> [(Indx,Indx)]
flipper [] = []
flipper ((c,h):hs) = h' ++ (flipper hs)
                         where h' = map (\x -> (x,c)) h


regrpHdr :: [(Indx,Indx)] -> Headers
regrpHdr hs =  map (\x -> (fst (head x),map snd x)) hs'
              where hs' = groupBy (\x y -> fst x == fst y) (sort hs)
--                    hs'' =  map (\x -> (fst (head x),map snd x)) hs'


(|||) :: (a -> Bool) -> (a -> Bool) -> a -> Bool
p ||| q = \x -> p x || q x

(&&&) :: (a -> Bool) -> (a -> Bool) -> a -> Bool
p &&& q = \x -> p x && q x

eqRows :: Indx -> Indx -> Bool
eqRows x y = row x == row y

eqCols :: Indx -> Indx -> Bool
eqCols x y = col x == col y
