--
-- Header.hs -- Header info
-- Owner -- Scott Andrew
--

module UCheck.Header where

--import Etc.Units
--import UCheck.Units
import Sheet.Indx
import Sheet.Sheet
import Sheet.Expr
import Maybe
import List
import UCheck.HdrDef
import UCheck.HdrAux

-- Different algorithms for header inference.
-- allHeaders creates a full set of possible headers - this is a simplistic approach
allHeaders :: Sheet Fml -> [(Indx,[Indx])]
allHeaders s = map (\x -> (fst x,filter (flip elem non) (snd x))) (filter (\x -> elem (fst x) non) all)
             where (Indx c1 r1) = minIndx s
                   (Indx c2 r2) = maxIndx s
		   all =  [(Indx x y,[Indx x y1 | y1 <- [1..y-1]]++
                              [Indx x1 y| x1 <- [1..x-1]])| x <- [c2,c2-1..c1], y <- [r2,r2-1..r1]]
		   non = findNonNulls s

-- findStrings builds a list of cells that have string values
-- these are candidates for headers.
-- no meu caso estas sao as candidatas a core
findStrings :: Sheet Fml -> [Indx]
findStrings s = map fst $ unSheet (rngFilter isString s) 

-- findInts builds a list of cells that have num values
-- these are candidates for core cells.
findInts :: Sheet Fml -> [Indx]
findInts s = map fst $ unSheet (rngFilter isNum s) 

-- findNonNulls finds the non-null cells 
findNonNulls :: Sheet Fml -> [Indx]
findNonNulls s = filter (\i -> isJust (s # i)) [(Indx x y)| x <- [c1..c2], y <- [r1..r2]]
	         where (Indx c1 r1) = minIndx s
                       (Indx c2 r2) = maxIndx s
		    
-- stringHeaders assumes only Strings are headers
stringHeaders :: Sheet Fml -> [(Indx,[Indx])]
stringHeaders s = map (\x -> (fst x,(filter (flip elem strings)) (snd x))) all
		  where all = allHeaders s
                        strings = findStrings s

stringHeaders' :: Sheet Fml -> Sheet Classif
stringHeaders' s = Sheet (zip (nub (concat (map (\x -> snd x) s'))) (repeat [(Hdr,2)])) 
                    where s' = stringHeaders s


-- intCore assumes all num values are core cells.
intCore :: Sheet Fml -> Sheet Classif
intCore s = Sheet (zip s' (repeat [(Core,2)]))
            where s' = findInts s


-- coreCell finds the core cells on the basis of formulae
coreCell :: Sheet Fml -> Sheet Classif
coreCell s | cs /= [] = Sheet (map (\x -> (x,[(Core,3)])) (foldr1 union cs))
           | otherwise = Sheet [] 
             where cs = map (\x -> filter (\y -> comp s (s # y) (s # x)) 
                        (x:(neighbor x))) (nub (concatMap (\x ->aggEx (snd x)) 
                        (unSheet s)))

-- fmlFooter classifies formula cells as footers.
-- this basically prevents them from being chosen as core.
fmlFooter :: Sheet Fml -> Sheet Classif
fmlFooter (Sheet s) = Sheet (map (\x -> (x,[(Ftr,4)])) fs')
             where fs = filter (\(i,f) -> isAgg f) s 
                   fs' = map (\x -> fst x) fs

-- fmlFooter' classifies binary formulas with references as footers.
-- this prevents them from being left unclassified.
fmlFooter' :: Sheet Fml -> Sheet Classif
fmlFooter' (Sheet s) = Sheet (map (\x -> (x,[(Ftr,3)])) fs')
              where fs = filter (\(i,f) -> isAgg' f) s 
                    fs' = map (\x -> fst x) fs


-- binFooter classifies formulas with binary operators as footers.
-- this prevents them from being left unclassified.
binFooter :: Sheet Fml -> Sheet Classif
binFooter (Sheet s) = Sheet (map (\x -> (x,[(Ftr,1)])) fs')
              where fs = filter (\(i,f) -> isBin f) s 
                    fs' = map (\x -> fst x) fs


-- borderChecker finds the cells that fall on the border
-- it marks data cells on top and left as headers and 
-- formula cells on bottom and right as footers.
--borderChecker :: Sheet Fml -> Sheet Classif


scorer :: Sheet Fml -> Sheet Classif
scorer (Sheet s) = Sheet ss'
                   where ss = map (\x -> scoreBuilder x (Sheet s)) s
                         ss' = map scoreClassif ss

scoreBuilder :: (Indx,Fml) -> Sheet Fml -> (Indx,Fml,Int,Int,Int,Int)
scoreBuilder (i,f) s = (i,f,tN,bN,lN,rN)
                      where tN = length (filter (\x -> isUnused x) (map (\x -> maybe (Inp Unused) id (s # x)) (topNeighbor i)))
                            bN = length (filter (\x -> isUnused x) (map (\x -> maybe (Inp Unused) id (s # x)) (botNeighbor i)))
                            lN = length (filter (\x -> isUnused x) (map (\x -> maybe (Inp Unused) id (s # x)) (leftNeighbor i)))
                            rN = length (filter (\x -> isUnused x) (map (\x -> maybe (Inp Unused) id (s # x)) (rightNeighbor i)))
 
scoreClassif :: (Indx,Fml,Int,Int,Int,Int) -> (Indx,Classif)
scoreClassif (i,f,ts,fs,ls,rs) | (ts>=2||ls>=2)&&(isString f) = (i,[(Hdr,ls+ts)])
                               | (ts>=2||ls>=2)&&(isAgg' f||isNum f) = (i,[(Hdr,ls+ts)])
                               | (fs>=2||rs>=2)&&(isAgg' f||isAgg f) = (i,[(Ftr,fs+rs)])
                               | (isUnused f) = (i,[(Fill,3)])
                               | otherwise = (i,[(Core,3)])

-- End of scorer algorithm

isSep :: Fml -> Bool
isSep = isUnused ||| isUnk ||| isBlank

{--MM
-- This was the first implementation of "hard fences" -  basically empty cells
-- that seperate tables. We try to identify the empty cells and align them to find 
-- they are seperators. We got the idea from image processing but it is pretty
-- expensive and complicated. The new implementation picks a seed and grows a region
-- till it can not grow anymore because the region is surrounded by empty cells. We 
-- then repeat this process till all the occupied cells in the worksheet have been
-- used up. In practice, this works better and we can "tune" it by changing our 
-- "neighbor" function.

hardFences :: Sheet Fml -> [(Indx,String)]
hardFences (Sheet cs) = top ++ bot ++ left ++ right
                        where top = concatMap (\x -> gTop (fst x) cs) cs
		              bot = concatMap (\x -> gBot (fst x) cs) cs
		              left = concatMap (\x -> gLeft (fst x) cs) cs
		              right = concatMap (\x -> gRight (fst x) cs) cs

gTop :: Indx -> [Cell Fml] -> [(Indx,String)]
gTop (Indx c r) cs | length fs >= 2 && not (isUnused cur) = [((Indx c r),"T")]
		   | otherwise = []
                     where ns = [(Indx x (r-1))|x <- [(c-1)..(c+1)]]
			   ns' = map (\x -> maybe (Inp Unused) id ((Sheet cs)#x)) ns
			   fs = filter isSep ns'
                           cur = maybe (Inp Unused) id ((Sheet cs)#(Indx c r))

gBot :: Indx -> [Cell Fml] -> [(Indx,String)]
gBot (Indx c r) cs | length fs >= 2 && not (isUnused cur) = [((Indx c r),"B")]
		   | otherwise = []
                     where ns = [(Indx x (r+1))|x <- [(c-1)..(c+1)]]
			   ns' = map (\x -> maybe (Inp Unused) id ((Sheet cs)#x)) ns
			   fs = filter isSep ns'
                           cur = maybe (Inp Unused) id ((Sheet cs)#(Indx c r))
                           
gLeft :: Indx -> [Cell Fml] -> [(Indx,String)]
gLeft (Indx c r) cs | length fs >= 2 && not (isUnused cur) = [((Indx c r),"L")]
		    | otherwise = []
                      where ns = [(Indx (c-1) y)|y <- [(r-1)..(r+1)]]
			    ns' = map (\x -> maybe (Inp Unused) id ((Sheet cs)#x)) ns
			    fs = filter isSep ns'
                            cur = maybe (Inp Unused) id ((Sheet cs)#(Indx c r))

gRight :: Indx -> [Cell Fml] -> [(Indx,String)]
gRight (Indx c r) cs | length fs >= 2 && not (isUnused cur) = [((Indx c r),"R")]
	 	     | otherwise = []
                       where ns = [(Indx (c+1) y)|y <- [(r-1)..(r+1)]]
			     ns' = map (\x -> maybe (Inp Unused) id ((Sheet cs)#x)) ns
			     fs = filter isSep ns'
                             cur = maybe (Inp Unused) id ((Sheet cs)#(Indx c r))

-- Functions for detecting "soft fences"
softFences :: Sheet Fml -> [[Indx]]
softFences (Sheet s) = colh++colf++rowh++ rowf
                       where fs = map fst (filter (\(i,f) -> isAgg f) s) 
			     hs = map fst (filter (\(i,f) -> isString f) s)
			     colf = filter ((>1).length) (groupBy eqCols fs)
			     colh = filter ((>1).length) (groupBy eqCols hs) 
			     rowf = filter ((>1).length) (groupBy eqRows fs) 
			     rowh = filter ((>1).length) (groupBy eqRows hs) 

--}


--MM New implementation of hard fences.
-- The following functions can be used to identify tables in worksheets.
-- We can detect hard fences if we use the hardSeeder and soft fences if 
-- we use the softSeeder.
-- Based on our experience, we don't think soft headers can be reliably 
-- inferred - so this should be used with caution. We are not using this 
-- right now and are working on figuring out more reliable approaches.

type Table = (Indx,Indx)  -- we assume rectangular tables
                          -- described by the diag. opp. corners

-- tables gives a list of tables as seperate worksheets
tables :: Sheet Fml -> [Sheet Fml]
tables s = map Sheet ts'
           where ss = sort (dom s)
		 ts = map sort (tabs s ([],ss))
		 ts' = map (map (\i -> (i,fromMaybe UnkFml (s # i)))) ts

-- tableRange gives a list of tuples denoting the upper left 
-- and bottom right indices of the identified tables
tableRange :: Sheet Fml -> [Table]
tableRange s = map (\x -> (head x,last x)) ts
           where ss = sort (dom s)
		 ts = map sort (tabs s ([],ss))		      

tabs :: Sheet Fml -> ([[Indx]],[Indx]) -> [[Indx]]
tabs s (xss,[]) = xss
tabs s (xss,y:ys) = tabs s (ns:xss,ns')
                      where (ns,ns') = seedStepper s 'H' y ([],ys)

type Flag = Char

seedStepper :: Sheet Fml -> Flag -> Indx -> ([Indx],[Indx]) -> ([Indx],[Indx])
seedStepper s f ix (xs,[]) = (ix:xs,[])
seedStepper s f ix (xs,ys) | xs++ns == [] = ([ix],ys)
		           | ns == [] = (ix:xs',ys')	    
                           | otherwise = seedStepper s f (head zs) (tail zs,ns')
                               where (ns,ns') = if f == 'H' then (hardSeeder ix ys) 
                                                else softSeeder s ix ys
                                     zs = xs++ns++[ix]
		         	     (xs',ys') = seedStepper s f (head xs) (tail xs,ys)    

--hardSeeder stops growing at hard fences.
hardSeeder :: Indx -> [Indx] -> ([Indx],[Indx])
hardSeeder ix xs = (ns,ns')
               where ns = delete ix (filter (isNeighbor ix) xs)
                     ns' = xs \\ (ix:ns) 

-- softSeeder stops growing at soft fences.
softSeeder :: Sheet Fml -> Indx -> [Indx] -> ([Indx],[Indx])
softSeeder s ix xs = (ns,ns')
                     where ns = delete ix (filter (\iy -> not (isSoftNeighbor s ix iy)) xs)
                           ns' = xs \\ (ix:ns) 


-- sheetMerge merges possibly overlapping sheets
sheetMerge :: (a -> a -> a) -> Sheet a -> Sheet a -> Sheet a
sheetMerge f (Sheet s) (Sheet s') = Sheet ss''
				     where ss = sortBy (\x y -> compare (fst x) (fst y)) (s ++ s')
 					   ss' = groupBy (\x y -> fst x == fst y) ss
					   ss'' = fApp f ss' 

fApp :: (a -> a -> a) -> [[Cell a]] -> [Cell a]
fApp f s = map (\x -> (fst (head x), foldr1 f (map (\y -> snd y) x))) s

--MM preHdrInf can be used to plug in and run all the classfier algorithms we have.
-- This makes it easy to add and try out new classifiers to make the header
-- inference more accurate.
preHdrInf :: Sheet Fml -> Sheet Classif
preHdrInf s = foldr1 (sheetMerge (++)) [s1,s2,s3,s4,s5,s6,s7]
              where s1 = stringHeaders' s 
                    s2 = coreCell s
                    s3 = fmlFooter s
                    s4 = intCore s
                    s5 = fmlFooter' s 
                    s6 = binFooter s
                    s7 = scorer s

-- conclude takes the evidence and assigns final classification
conclude :: (Indx,Classif) -> (Indx,CellRole)
conclude (i,cs) | h > c && h > f = (i,Hdr)
                | c > h && c > f = (i,Core)
                | otherwise = (i,Ftr)
		 where cs' = map (\f -> filter f cs) (map (extr fst) [Hdr,Core,Ftr])
                       [h,c,f] = map (\xs -> sum (map snd xs)) cs'

extr :: Eq a => (t -> a) -> a -> t -> Bool
extr f role = \x -> f x == role

--MM regions breaks up the sheet on the basis of roles
-- basically into header and non-header (core+footer) regions
regions :: Sheet Classif -> ([Indx],[Indx])
regions (Sheet s) = (f++c,h)
		   where roles = map conclude s
			 roles' = map (\f -> filter f roles) (map (extr snd) [Hdr,Core,Ftr])
			 [h,c,f] = map (map fst) roles'

-- this version of hdrInf picks the nearest header cells to the left and above target cell.
hdrInf :: ([Indx],[Indx]) -> Headers
hdrInf ([],_) = []
hdrInf (c:cs,hs) = (c,rh++ch):(hdrInf (cs,hs))
                   where rowHdrs = filter (eqRows c) hs 
                         colHdrs = filter (eqCols c) hs 
                         rowds =  map (\x -> ((colDist c x),x)) rowHdrs
                         rowds' = filter (\(x,y) -> x > 0) rowds
 			 colds =  map (\x -> ((rowDist c x),x)) colHdrs
                         colds' = filter (\(x,y) -> x > 0) colds
                         rh = if rowds' /= [] then [snd (minimumBy compareDist rowds')] else []
                         ch = if colds' /= [] then [snd (minimumBy compareDist colds')] else []
			 


-- hdrHdr generates the list of headers of headers.
-- first parameter is list of headers that have been assigned
-- second parameter is list if unassigned headers.
-- third parameter is the Header information already built
hdrHdr :: [Indx] -> [Indx] -> Headers
hdrHdr as us = hdrHInf (as,us)

--MM as is the list of assigned header, and us is the list of unassigned headers. 
-- Indices in us can be higher-level row or column headers - we need to consider both.
-- we don't consider 3rd and higher level headers in this version - so hdrHInf 
-- does not call itself recursively. Moreover, inference of different table reagions in
-- the same worksheet is also a little broken since we are not using the soft fences
-- approach.
hdrHInf :: ([Indx],[Indx]) -> Headers
hdrHInf (as,us) = opt
                  where byRowAs = sort (byRow as)
                        byColAs = sort (byCol as)
                        byRowUs = (reverse.sort) (byRow us)
                        byColUs = (reverse.sort) (byCol us)
                        rowAsg  = rowAssign byRowAs byRowUs
                        colAsg  = colAssign byColAs byColUs
                        opt = optAlg2 (rowAsg++colAsg)

{-
-- ** testing code
rgs = regions (preHdrInf harvest)
hs = hdrInf rgs
asg = nub (concatMap (\x -> snd x) hs)
unAsg = (snd rgs) \\ asg
byRowAs = sort (byRow asg)
byColAs = sort (byCol asg)
byRowUs = (reverse.sort) (byRow unAsg)
byColUs = (reverse.sort) (byCol unAsg)
rowAsg = rowAssign byRowAs byRowUs
colAsg = colAssign byColAs byColUs

-- ** end of testing code
-}

optAlg2 :: [(Indx,[Indx],Cost)] -> Headers
optAlg2 [] = []
optAlg2 as  = hs ++ (optAlg2 as')
                where (i,is,mc) = minimumBy (\(x,y,z) (x',y',z') -> compare z z') as
		      hs = map (\x -> (x,[i])) is
		      rs' = filter (\(x,y,z) -> x /= i) as
		      as' = filter (\(x,y,z) -> y /= is) rs'

-- rowAssign returns assigned headers, unassigned headers and a 3-tuple:
-- (h,hs,Cost): Cost of assigning h as headers to hs
rowAssign :: [(Row,[Indx])] -> [(Row,[Indx])] -> [(Indx,[Indx],Cost)] 
rowAssign as [] = []
rowAssign as (u:us) | a' /= [] = ps++(rowAssign (as++[(r,na)]) us) 
                    | otherwise = rowAssign as us
                     where a' = filter (\x -> fst x >= fst u) as
                           a = head a'
                           r = fst u
                           ps = processRow (snd a) (snd u)  
                           na = map (\(n,ns,c) -> n) ps
                           ns = map (\(n,ns,c) -> ns) ps
                           as' = as \\ (byRow ((nub.concat) ns))

processRow :: [Indx] -> [Indx] -> [(Indx,[Indx],Cost)]
processRow [] _ = []
processRow as [u]= [(u,as,(rowCost u as))]
processRow (a:as) (u1:u2:us) = (u1, as1, (rowCost u1 as1)):(processRow as' (u2:us))
			       where d = -1 * (colDist u1 u2)
                                     c = col a
                                     as1 = filter (\x -> col x >= c && col x < (c+d)) (a:as)
                                     as' = (a:as) \\ as1



-- colAssign returns assigned headers, unassigned headers and a 3-tuple:
-- (h,hs,Cost): Cost of assigning h as headers to hs
colAssign :: [(Col,[Indx])] -> [(Col,[Indx])] -> [(Indx,[Indx],Cost)] 
colAssign as [] = []
colAssign as (u:us) | a' /= [] = ps++(colAssign (as++[(r,na)]) us) 
                    | otherwise = colAssign as us
                     where a' = filter (\x -> fst x >= fst u) as
                           a  = head a'
                           r = fst u
                           ps = processCol (snd a) (snd u)
                           na = map (\(n,ns,c) -> n) ps
                           ns = map (\(n,ns,c) -> ns) ps
                           as' = as \\ (byCol ((nub.concat) ns))

processCol :: [Indx] -> [Indx] -> [(Indx,[Indx],Cost)]
processCol [] _ = []
processCol as [u] = [(u,as,(colCost u as))]
processCol (a:as) (u1:u2:us) = (u1, as1, (colCost u1 as1)):(processCol as' (u2:us))
			       where d = -1 * (rowDist u1 u2)
                                     r = row a
                                     as1 = filter (\x -> row x >= r && row x < (r+d)) (a:as)
                                     as' = (a:as) \\ as1


-- allRun is just a wrapper to make the call easier
allRun :: Sheet Fml -> Headers
allRun (Sheet []) = []
allRun s = oneHeader (hs ++ hss)
           where rgs = regions (preHdrInf s)
                 hs  = hdrInf rgs
                 asg = nub (concatMap (\x -> snd x) hs)
                 unAsg = (snd rgs) \\ asg
                 hss  = hdrHdr asg unAsg


-- oneHeader removes the cases of headers that have only a single child
oneHeader :: Headers -> Headers
oneHeader hs = flipHeader hs''
               where hs' = flipHeader hs
                     hs'' = filter (\(x,y)->length y > 1) hs'

