--
-- Units.hs -- Units (inferring and checking for well-formed)
-- Owner -- Scott Andrew
--

{-

   --------------------------------------------------
   -- notes
   --------------------------------------------------

   - 

-}

module UCheck.Units where

--------------------------------------------------
-- imports
--------------------------------------------------

import Sheet.Expr
import Sheet.Indx
import Sheet.Sheet
import UCheck.HdrAux
import UCheck.HdrDef
import UCheck.UnitErrors
import Util.PrintList

import List
import Maybe
import Debug.Trace

--------------------------------------------------
-- unit types
--------------------------------------------------

type Base = String
type DepUnit = [Base]
type AndUnit = [DepUnit]
type OrUnit = [AndUnit]

oneDep = []
oneAnd = [oneDep]
oneOr = [oneAnd]
oneUnit = U oneOr

newtype Unit = U {unU :: OrUnit}
        deriving (Ord, Eq)

type SheetU = Sheet Unit

--------------------------------------------------
-- making AND/OR units
--------------------------------------------------

-- makes an AND unit from the 2 passed units
--
mA :: Unit -> Unit -> Unit
mA (U as) (U bs) = U $ map nub $ productWith (++) as bs

-- makes an OR unit from the passed list of units
--
mO :: [Unit] -> Unit
mO us = U $ nub $ concatMap unU us

productWith :: (a -> b -> c) -> [a] -> [b] -> [c]
productWith f xs ys = [f x y | x <- xs, y <- ys]

--------------------------------------------------
-- unit inference
--------------------------------------------------

inferU :: Headers -> Sheet Fml -> SheetU
inferU hs (Sheet cs) = -- trace "inferU"
  Sheet (map (\(a,c)->(a, inferUC cs hs (a,c))) cs)

{-

To infer the unit for a cell a that contains a reference to cell b,
we consider the following cases:

(1) Cell a has a header (lookup a hs = Just v)
    (1.1) Cell a contains just the reference to b (lookup a cs == Just (Ref b))
          (1.1.1) Cell b exists (lookup b cs = Just k)
                  (This is the case for the reference unit rule, which takes
                   the AND unit of a and b.)

                  => mA (getUnit cs hs a) (inferUC cs hs (b,k))

          (1.1.2) Cell b does not exist (lookup b cs = Nothing)
                  (This case occurs when a reference points outside of the
                   considered spreadsheet area, in which case we can simply
                   take the unit of cell a.)

                  => getUnit cs hs a

    (1.2) Cell a contains a formula different from Ref b
          (This case occurs when inferUC is called recursively for
           arguments of a formula. In these recursive calls the address
           of the formula is passed as the cell index. In this case the
           "strong" reference rule does not apply, and we simply take
           the unit of cell b.)

           => getUnit cs hs b

(2) Cell a has no header (lookup a hs = Nothing)
    (2.1) Cell b exists (lookup b cs = Just k)
          take the header of b as the unit (inferUC cs hs (b,k))

    (2.2) Cell b does not exist (lookup b cs = Nothing)

          => oneUnit

-}

inferUC :: [Cell Fml] -> Headers -> Cell Fml -> Unit

inferUC cs hs (a,Inp v) = -- trace ("Inp: " ++ show (a, Inp v)) 
                          getUnit cs hs a

{-

first check the header info
no header then check Ref
"yes" header, then And units of header and Ref cell units

-}

inferUC cs hs (a,Ref b) =
 
   case lookup a hs of --(trace ("Ref: " ++ show (a, Ref b)) hs) of
        Just v -> if lookup a cs == Just (Ref b) then
                     case lookup b cs of
                          Just k -> mA (getUnit cs hs a) (inferUC cs hs (b,k))
                          _      -> getUnit cs hs a
                  else
                     case lookup b cs of
                          Just k -> inferUC cs hs (b,k)
                          _      -> oneUnit
        _      -> maybe (U [[[]]]) (\k->inferUC cs hs (b,k)) (lookup b cs)

{-

code for handling case of Binary operations
we might have the cases:
1) = 2 + 3   - simple mathematical operation
2) = B2 + B3 - math operation involving references
3) = 2 * B3 - mixture of 1 and 2
we adopt same strategy as for aggregation

-}

inferUC cs hs (a,Bin o f1 f2) =
 -- trace ("Bin: " ++ show (a, Bin o f1 f2))
   mO [(inferUC cs hs (a,f1)), (inferUC cs hs (a,f2))]

-- [scott] added to handle (=sum(d3:d5)) cases from within excel
inferUC cs hs (a,RefRng xs) =
 --trace ("RefRng: " ++ show (a, RefRng xs))
   mO ((map (inferUC cs hs) (map ((,)a) xs)))

inferUC cs hs (a,Agg Plus xs) =
trace ("Agg1: " ++ show (a, Agg Plus xs))
   mO ((map (inferUC cs hs) (map ((,)a) xs)))

inferUC cs hs (a,Agg Mult (x:xs))
   |unit1 /= -- trace ("Agg2: " ++ show (a, Agg Mult (x:xs))) 
             U [[[]]] && all(== oneUnit)(map (inferUC cs hs) (map ((,)a) xs))
	   = inferUC cs hs (a,x)
   |unit1 == -- trace ("Agg2: " ++ show (a, Agg Mult (x:xs)))
              U [[[]]] = inferUC cs hs (a,Agg Mult xs)
   |otherwise = --  trace ("Agg2: " ++ show (a, Agg Mult (x:xs))) 
              oneUnit
                where unit1 = inferUC cs hs (a,Agg Mult xs)

inferUC cs hs (a,Agg f (x:xs)) =
   case x of -- (trace ("Agg3: " ++ show (a, Agg f (x:xs))) x) of
        Ref b -> mA (inferUC cs hs (a, Ref b)) (getUnit cs hs a)
        _     -> oneUnit

-- [scott] added to handle the "=" start of a formula within excel
inferUC cs hs (a,Outer o) = -- trace ("outer: " ++ show (a, Outer o)) 
  inferUC cs hs (a,o)

inferUC _ _ c = -- trace ("OneUnit: " ++ show c)
  oneUnit

--------------------------------------------------
-- unit tracing code
--------------------------------------------------

-- some standard unexpected errors
--
cellNotFound x = error ("Cell "++show x++" does not exist.")
tooManyHdrs a xs = error ("Unexpected error [traceUnit]: Cell "
                           ++show a++
                          " contains zero or more than one headers, namely: "
                           ++show xs)

{-

construct unit for input cells from headers

The unit for a cell (a,v) in the context of headers hs is either:

1   if there is no (a,xs) in hs

or:

u_1 & ... & u_n   when (a,[x_1,...,x_n]) is in hs and
u_i = [a_1,a_2,...,a_k1]
   and a_1 = cs!!x_i
      a_2 = cs!!y  where (x_i,[y]) in hs
      a_3 = cs!!z  where (y,[z]) in hs
      ...

-}

getUnit :: [Cell Fml] -> Headers -> Indx -> Unit
getUnit cs hs a = case lookup a hs of
                     Nothing -> oneUnit
                     Just xs -> U [map (traceUnit cs hs) xs]

getCell :: Indx -> [Cell Fml] -> String
getCell x cs = case lookup x cs of
                    Just (Inp (S s)) -> s
                    _                -> cellNotFound x

-- note: lists representing dependent units are from general to specific
--
traceUnit :: [Cell Fml] -> Headers -> Indx -> DepUnit
traceUnit cs hs a = case lookup a hs of
                         Nothing  -> [getCell a cs]
                         Just []  -> [getCell a cs]
                         Just [x] -> traceUnit cs hs x++[getCell a cs]
                         Just xs  -> tooManyHdrs a xs

--------------------------------------------------
-- unit checking
--------------------------------------------------

unitChecker :: SheetU -> SheetU -> Headers -> Sheet Fml -> ErrRpt
unitChecker (Sheet []) _ _ _ = []
unitChecker sht@(Sheet ((i,u):us)) s hs (Sheet fs) = if wfuRes /= NoErr then
                                                     (i,[wfuRes]) : (unitChecker (Sheet us) s hs (Sheet fs)) -- [scott] made ErrDesc into [ErrDesc] here
                                                   else
                                                     unitChecker (Sheet us) s hs (Sheet fs)
                                                   where wfuRes = wfu s hs fs u

-- Determines if a unit is well-formed
--
wfu :: SheetU -> Headers -> [Cell Fml] -> Unit -> ErrDesc
wfu _ _ _ (U [[[_]]]) = NoErr
wfu _ _ _ (U [ds]) = if (validAnd ds) then
                       NoErr
                     else
                       UnkErr
wfu (Sheet us) hs fs (U as) = validOr us hs fs as

validAnd :: AndUnit -> Bool
validAnd [d] = True
validAnd ds@(_:_:_) = unique level && validAnd rest
                      where ds'   = filter (not . null) ds
                            level = map head ds'
                            rest  = map tail ds'
validAnd _ = True

unique :: Eq a => [a] -> Bool
unique xs = length (nub xs) == length xs

sameDims :: [[a]] -> Bool
sameDims as = allEq us
              where us = map length as

allEq :: Eq a => [a] -> Bool
allEq xs = length (nub xs) == 1

sameDeps :: OrUnit -> Bool
sameDeps ([]:_) = True
sameDeps as = sameDep (map head as) && sameDeps (map tail as)

sameDep :: AndUnit -> Bool
sameDep ds = sameDims ds && allEq ds'
             where ds' = map init ds

--------------------------------------------------
-- new validOr implementation
--------------------------------------------------

{-

Algorithm: Valid OR Unit

IDEA:   Process all and units to maintain a space (s) spanned
         by a list of unit lists and a set of points (ps), each given
         by a list of base units

INPUT:  An or unit (i.e. a list of and units) bound to as

OUTPUT: List of missing units (given by points s \\ ps)
         as is a valid OR unit <=> points s \\ ps == []

METHOD:

1. Determine the "dimensionality" of the space.
    Each dimension is identified by the init of each
    dependent unit.

2. For each and unit a in as do:
      extend the space and the point set, i.e.:
        s := extendSpace p s
        ps := addPoint p ps
      where ps = map last ds

3. return:
    (a) points s \\ ps         (missingUnits)
    (b) points s \\ ps == []   (validOr)

-}

-- dimensions and space
--
type UDim  = [DepUnit]
type Space = [UDim]

emptySpace :: OrUnit -> Space
emptySpace as = replicate (maximum (map length as)) []

extendDim :: DepUnit -> UDim -> UDim
extendDim b bs | b `elem` bs = bs
               | otherwise   = b:bs

extendSpace :: Point -> Space -> Space
extendSpace = zipWith extendDim

points :: Space -> PointSet
points []     = []
points [d]    = map (:[]) d
points (d:ds) = [x:p | x <- d, p <- points ds]

-- points and point sets
--
type Point = AndUnit
type PointSet = [Point]

addPoint :: Point -> PointSet -> PointSet
addPoint = (:)

-- processing AND units to create spaces and point sets
--
type QuasiSpace = (Space,PointSet)

procAnd :: AndUnit -> QuasiSpace -> QuasiSpace
procAnd ds (s,ps) = (extendSpace bs s,addPoint bs ps)
                     where bs = filter (not.null) ds

space :: OrUnit -> QuasiSpace
space as = foldr procAnd (emptySpace as,[]) as

-- compare list of cell units (orunits) to find if there is a unit errors
-- cell unit = (indx unit orunit)
-- type Headers = [(Indx,[Indx])]
-- cell fml = (indx fml)
validOr :: [Cell Unit] -> Headers -> [Cell Fml] -> OrUnit -> ErrDesc
validOr us hs fs ou = case missUnts of
                         []  ->  NoErr -- set of missing and units was empty
                         xs  ->  if null missIndxsUnts then -- what does the arrow mean?
                                    ShrnkRng (getOddIndxs (getRefCntsFromFmlLst fs) (getRefsFromFml (getFmlFromIndx fs (getIndxFromUnit us (U ou)))))
                                 else
                                    getFllRngErr srcErrRefsUnts (getFllRngLst srcErrRefsUnts missIndxsUnts)
                                    --ScottTmp missUnts
                                 where srcErrIndx = getIndxFromUnit us (U ou)
                                       srcErrRefs = getRefsFromFml (getFmlFromIndx fs srcErrIndx)
                                       srcErrRefsUnts = map (\x->(x,unU (getUnitFromIndx us x))) srcErrRefs
                      where missUnts = missingUnits ou -- list of andunits that need to be included in problem orunit to make it valid	
                            missIndxsUnts = map (\x->(getIndxFromUnit us (U [x]),x)) (concatMap f missUnts) -- list of tuples with index and the andunit for each that needs to be included in orunit to make it valid...ensures getIndxFromUnit is only called with units that actually exist in the sheet rather than units that don't exist because they are a reference to another unit
                            f x = if (U [x]) `elem` (map snd us) then [x] else []

missingUnits :: OrUnit -> [Point]
missingUnits as = points s \\ ps
                   where (s,ps) = space as

--------------------------------------------------
-- applying the change suggestions
--------------------------------------------------

remRefFromIndx :: [Cell Fml] -> Indx -> Indx -> [Cell Fml]
remRefFromIndx fs srcix remix = (delete (srcix,srcfml) fs) ++ [(srcix,(remRefFromFml srcfml remix))]
                                where srcfml = getFmlFromIndx fs srcix

remRefFromFml :: Fml -> Indx -> Fml
remRefFromFml (RefRng rng) ix = RefRng (delete (Ref ix) rng)
remRefFromFml (Agg op rng) ix = Agg op (delete (Ref ix) rng)

addRefToIndx :: [Cell Fml] -> Indx -> Indx -> [Cell Fml]
addRefToIndx fs srcix addix = (delete (srcix,srcfml) fs) ++ [(srcix,(addRefToFml srcfml addix))]
                              where srcfml = getFmlFromIndx fs srcix

addRefToFml :: Fml -> Indx -> Fml
addRefToFml (RefRng rng) ix = RefRng (rng ++ [Ref ix])
addRefToFml (Agg op rng) ix = Agg op (rng ++ [Ref ix])

--------------------------------------------------
-- fill range algorithm
--------------------------------------------------

getFllRngErr :: [(Indx,OrUnit)] -> [(Indx,Indx)] -> ErrDesc
-- takes first variable of both arguments. forms set where they are the same
-- if that set is the same length as the length of is (first argument) then fillrng error
getFllRngErr refIs is = if length (groupBy (\(i,_) (i',_)->i == i') (sort is)) == length is then
                           FillRng (map (\x->(head (fst (unzip x)),snd (unzip x))) (groupBy (\(i,_) (i',_)->i == i') (sort (map (\(i,i')->(i',i)) is))))
                        else
                           UnkFill

cntSimUnts :: AndUnit -> OrUnit -> [Bool]
cntSimUnts au aus = map (\au'->isVldFll [au'] au) aus

-- returns a list of tuples indicating in each that the first index needs to be included as a reference in the second
getFllRngLst :: [(Indx,OrUnit)] -> [(Indx,AndUnit)] -> [(Indx,Indx)]
getFllRngLst aggUnts missUnts = concatMap (\(i,au)->concatMap (\(i',ou)->if isVldFll ou au then [(i,i')] else []) aggUnts) missUnts

-- returns true if the andunit is a valid fit to fill the orunit and false otherwise
isVldFll :: OrUnit -> AndUnit -> Bool
isVldFll ou au = if length ou > 1 then
                    intersect au ouCmnDepUnts == ouCmnDepUnts -- true if the andunit contains in it the same common depunits as the orunit has and false otherwise
                 else
                    isctLen > 0 && auLen - isctLen == 1 -- true if the andunit has all common depunits as the one andunit in the orunit except for one and false otherwise
                 where ouCmnDepUnts = getCmnDepUnts ou
                       isctLen = length (intersect (head ou) au)
                       auLen = length au

-- returns all of the common depunits between each andunit in the orunit
-- if there is only one andunit in the orunit then that andunit is returned
getCmnDepUnts :: OrUnit -> [DepUnit]
getCmnDepUnts [] = []
getCmnDepUnts [au] = au
getCmnDepUnts (au:aus) = intersect au (getCmnDepUnts aus)

{- tests for use within unitstest
st = getFllRngErr (getFllRngLst seru miu)
-}

--------------------------------------------------
-- shrink range algorithm
--------------------------------------------------

-- returns the odd indices by looking at the number of references for each index in arg2 and grouping to find the individuals that have no commonalities
getOddIndxs :: [(Indx,Int)] -> [Indx] -> [Indx]
getOddIndxs allIxs curIxs = map (\(x,y)->x) (concat (filter (\x->length x == 1) (groupBy (\(_,y) (_,y')->y == y') (map (\(x,y)->(y,x)) (sort (map (\(x,y)->(y,x)) (getTheseRefCnts allIxs curIxs)))))))

{- tests for use within unitstest
st = getRefCntsFromFmlLst (dropS rngTooLrgErrSheet)
st2 = getRefsFromFmlLst [(c 6, agg Plus [c 2, c 3, c 4, c 5])]
st3 = getTheseRefCnts st st2
st4 = getOddIndxs st st2
-}

--------------------------------------------------
-- scott's utility functions
--------------------------------------------------

-- searches arg1 for arg2 and returns the index
getIndxFromUnit :: [Cell Unit] -> Unit -> Indx
getIndxFromUnit us u = fromJust (lookup u [(u,i) | (i,u) <- us])

-- searches arg1 for arg2 and returns the formula
getFmlFromIndx :: [Cell Fml] -> Indx -> Fml
getFmlFromIndx fs ix = fromJust (lookup ix fs)

-- searches arg1 for arg2 and returns the unit
getUnitFromIndx :: [Cell Unit] -> Indx -> Unit
getUnitFromIndx us ix = fromJust (lookup ix us)

-- returns all of the indices immediately referenced by the passed formula
getRefsFromFml :: Fml -> [Indx]
getRefsFromFml (Ref ix) = [ix]
getRefsFromFml (RefRng rng) = concatMap getRefsFromFml rng
getRefsFromFml (Agg _ rng) = concatMap getRefsFromFml rng
getRefsFromFml _ = []

-- returns all of the indices immediately referenced by looking at each formula in arg1
getRefsFromFmlLst :: [Cell Fml] -> [Indx]
getRefsFromFmlLst [] = []
getRefsFromFmlLst ((_,f):fs) = (getRefsFromFml f) ++ (getRefsFromFmlLst fs)

-- returns a list of tuples holding each index referenced within arg1 and how many times it was
getRefCntsFromFmlLst :: [Cell Fml] -> [(Indx,Int)]
getRefCntsFromFmlLst fs = map (\x->(head x,length x)) (group (sort (getRefsFromFmlLst fs)))

-- returns a list of tuples with the indices from arg2 and how many times they are each referenced within arg1
getTheseRefCnts :: [(Indx,Int)] -> [Indx] -> [(Indx,Int)]
getTheseRefCnts allIxs curIxs = concatMap (\(x,y)->if (intersect [x] curIxs) /= [] then [(x,y)] else []) allIxs

-- drops the sheet constructor to return a list of cell types
drpSht :: Sheet a -> [Cell a]
drpSht (Sheet xs) = xs

--------------------------------------------------
-- show units info
--------------------------------------------------

instance Show Unit where
	show(U [[[]]]) = "1"
        show(U u) = printOr u

printOr  =  printList[""," | ",""] printAnd
printAnd =  printList["","&",""] printDep
printDep =  printList["",".",""] id
