{-# OPTIONS -fglasgow-exts #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Point-free variation of TwoLevel.hs
--
-----------------------------------------------------------------------------

module Data.Transform.TwoLevelPF (
    module Data.Transform.TwoLevelPF,
    module Data.Transform.Rule,
    module Data.Transform.Constraints
 ) where

import Data.Set as Set hiding (split)
import Data.Map as Map hiding (split)
import Control.Monad.State
import Data.Maybe
import Data.List as List
import Debug.Trace

import Data.Transform.Rule hiding ((/\))
import Data.Transform.RulesPF (optimize, simplifyPF, simplifyRel, simplifyrdb)
import Data.Transform.Rewriting (reduce, everywhere)
import Data.Transform.Constraints
    
-----------------------------------------------------------------------------
-- * Strategy combinators for two-level transformations

-- | Sequential composition.
(>>>) :: Rule -> Rule -> Rule
(f >>> g) a = do 
  View r b <- f a
  View s c <- g b
  return (View (comprep b r s) c)

-- | Left-biased choice.
(|||) :: Rule -> Rule -> Rule
(f ||| g) x = f x `mplus` g x

-- | Do-nothing operation. 
--   Unit for sequential composition. 
--   Left-zero and right-unit for left-biased choice.
nop :: Rule
nop x = Just (View idrep x)

-- | Repeat until failure.
--   Zero or more repetitions.
many :: Rule -> Rule
many r = (r >>> many r) ||| nop

-- | Apply argument rule exactly once, at arbitrary depth.
once :: Rule -> Rule
once r Int  = r Int
once r Bool = r Bool
once r Char = r Char
once r One  = r One
once r String = r String
once r (List a) = r (List a) `mplus` 
                  (do View s b <- once r a
                      return (View (listmaprep s) (List b)))
once r (Set a) = r (Set a) `mplus` 
                 (do View s b <- once r a
                     return (View (setmaprep s) (Set b)))
once r (Map a b) = r (Map a b) `mplus`
                   (do View s c <- once r a
                       return (View (mapkeysrep s) (Map c b))) `mplus`
                   (do View s c <- once r b
                       return (View (mapmaprep s) (Map a c)))
once r (Either a b) = r (Either a b) `mplus`
                      (do View s c <- once r a
                          return (View (summaprep s idrep) (Either c b))) `mplus`
                      (do View s c <- once r b
                          return (View (summaprep idrep s) (Either a c)))
once r (Prod a b) = r (Prod a b) `mplus`
                    (do View s c <- once r a
                        return (View (prodmaprep s idrep) (Prod c b))) `mplus`
                    (do View s c <- once r b
                        return (View (prodmaprep idrep s) (Prod a c)))
once r (Tag n a) = r (Tag n a) `mplus` 
                   (do View s b <- once r a
                       return (View s (Tag n b)))

once r (TInv t i)
   = r (TInv t i) `mplus` 
     (do View s b <- once r t
         let inv' = {-trySimplify (Func b Bool) $ -} COMP t i (from s)
         return $ View s (TInv b inv')
     )
once r x = r x

-- | Apply argument rule exactly once, at arbitrary depth traversing only
--   top-level TInv and Prod definitions
onceTL :: Rule -> Rule
onceTL r (Prod a b)
   = (r (Prod a b)) `mplus`
     (do View s c <- onceTL r a
         return $ View (reppi1 s) (Prod c b)) `mplus`
     (do View s c <- onceTL r b
         return $ View (reppi2 s) (Prod a c))
   where reppi1 s = Rep { from = PROD (from s) ID, to   = PROD (to s) ID }
         reppi2 s = Rep { from = PROD ID (from s), to   = PROD ID (to s) }
onceTL r (TInv t i)
   = r (TInv t i) `mplus`
     (do View s b <- onceTL r t
         let inv' = COMP t i (from s)
         return $ View s (TInv b inv'))
onceTL _ _ = mzero


or_propagate :: Rule -> Rule
or_propagate r = r ||| propagate r

ors :: [Rule] -> Rule
ors [] = stop
ors [r] = r
ors (r:rs) = r ||| ors rs

stop :: Rule
stop t = mzero

-----------------------------------------------------------------------------
-- * Rules for format evolution

----------------------------------------
-- ** Enrichment and removal

-- | Add a field (to the right).
--   The forward value-level function inserts a 
--   constant value for the new field.
addfield :: Type b -> b -> Rule
addfield b x a = return $ View rep (Prod a b)
    where rep = Rep (SPLIT ID (pnt x)) FST

pnt x = COMP One (PNT x) BANG


-- | Add a field (to the left).
--  The forward value-level function inserts a constant value for the new field.
addfieldl :: Type b -> b -> Rule
addfieldl b x a = return $ View rep (Prod b a)
  where rep = Rep (SPLIT (pnt x) ID) SND


-- | Type of generic value-level queries.
type Query b = forall a . Type a -> PF (a -> b)

-- | Add a field.
--   The forward value-level function inserts a field
--   determined by a query on its input value.
addinfo :: Type b -> Query b -> Rule
addinfo b f a = return $ View rep (Prod a b)
  where
    rep = Rep (SPLIT ID (f a)) FST
    
----------------------------------------
-- ** Generalization and restriction

-- | Add alternative.
--   The backward value level function fails if the added 
--   alternative is present.
addalt :: Type b -> Rule
addalt b a = return $ View rep (Either a b)
    where rep = Rep INL (EITHER ID (pnt (error "undefined #9")))

-- | Allow repetition.
--   Maps optional to a list.
allowRep :: Rule
allowRep (Either a One) = return $ View (Rep opt2seq seq2opt) (List a)
   where
     opt2seq = EITHER (wrap a) NIL
     seq2opt = FUN "seq2opt" (\l -> if (List.null l) then Right () else Left (head l)) 

wrap :: Type a -> PF (a -> [a])
wrap a = COMP (Prod a (List a)) CONS (SPLIT ID (COMP One NIL BANG))

----------------------------------------
-- ** Focus on tags

-- | Perform the argument rule inside the given tag.
inside :: String -> Rule -> Rule
inside n r (Tag m a) | n==m = do View r b <- r a
                                 return $ View r (Tag m b)
inside _ _ _ = mzero

-- | Perform the argument rule on the given tag.
atTag :: String -> Rule -> Rule
atTag n r x@(Tag m a) | n==m = r x
atTag _ _ _ = mzero

-----------------------------------------------------------------------------
-- * Rules for data mapping

-- | Flatten a hierarchical data type.
--   Does not include the rule for recursion elimination.
flatten :: Rule
flatten = many (aux >>> many aux >>> prods) >>> many (once assocr)
  where
    aux = once (ors [listmap, mapsum, summap, mapprodmap, setmap, prodsum, sumone, prodone])

{-
-- | Map a hierarchical datatype to a relational type.
toRDB :: Rule
toRDB = many (once fixastable) >>> many (once enum2int) >>> removetags >>> flatten
-}

-- | Flatten a hierarchical data type. Constraint-aware.
flattenInv :: Rule
flattenInv = many (aux >>> many aux >>> prods) >>> pullAssocR >>> fuse_constraint -- >>> once fuse_constraint
  where
    aux = once (ors (List.map or_propagate rules))
    rules :: [Rule]
    rules = [listmap, mapsum', summap, split, setmap, prodsum, sumone, prodone]
    
-- | Associate products to right and pull invariant definition to the top level
pullAssocR :: Rule
pullAssocR = many (onceTL (assocr ||| compose_constraint))
   where assocr :: Rule
         assocr (Prod (Prod a b) c) = return $ View rep (Prod a (Prod b c))
             where
               rep = Rep { from = assoclPF b c, to = assocrPF a b }
         assocr (Prod (TInv (Prod a b) i) c)
            = return $ View rep (TInv (Prod a (Prod b c)) i')
              where
                rep = Rep { from = assoclPF b c, to = assocrPF a b }
                i'  = COMP (Prod a b) i (PROD ID FST) 
         assocr (Prod a (TInv (Prod b c) i))
            = return $ View idrep (TInv (Prod a (Prod b c)) i')
              where 
                i' = COMP (Prod b c) i SND
         assocr _ = fail "Cannot apply!"

----------------------------------------
-- ** Elimination and introduction

-- | Eliminate list by refinement to map.
listmap :: Rule
listmap (List a) = return $ View (Rep LIST2MAP ELEMS) (Map Int a)
listmap _ = mzero

-- | Eliminate set by refinement to map.
setmap :: Rule
setmap (Set a) = return $ View (Rep (FUN "set2map" set2fm) DOM) (Map a One) 
  where 
    set2fm s = Map.fromDistinctAscList (zip (Set.toAscList s) (repeat ()))
setmap _ = mzero

-- | Eliminate optional by refinement to map.
sumone :: Rule
sumone (Either a One) 
  = return $ View (Rep optelim (FUN "optintro" optintro)) (Map One a)
  where
    optelim = EITHER (FUN "(Map.singleton ())" (Map.singleton ())) EMPTY
    optintro m | Map.member () m = let Just x = Map.lookup () m in Left x
	 	      | otherwise = Right ()
sumone _ = fail "Not applicable"



----------------------------------------
-- ** Distribution

-- | Distribute map over sum.
mapsum :: Rule
mapsum (Map a (Either b c))
   = return $ View (Rep (FUN "uncojoin" uncojoin) (FUN "cojoin" (cojoin a))) (Prod (Map a b) (Map a c))
  where
    uncojoin :: Map a (Either b c) -> (Map a b, Map a c)
    uncojoin = (  fmap (either id (error "undefined #1")) . Map.fromDistinctAscList >< 
                  fmap (either (error "undefined #2") id) . Map.fromDistinctAscList) .
               List.partition (isLeft . snd) . Map.toAscList
    -- Pre: non duplicate keys
    cojoin :: Type a -> (Map a b, Map a c) -> Map a (Either b c)
    cojoin t = Map.fromDistinctAscList . 
               sortBy (\x y -> gcompare t (fst x) (fst y)) .
               uncurry (++) . 
               (Map.toAscList . fmap Left >< Map.toAscList . fmap Right)
mapsum _ = mzero

mapsum' :: Rule
mapsum' (Map a (Either c d)) = return $ View rep (TInv t (mapsum_inv a One c d))
   where rep = Rep { to = (error "undefined #3"), from = (error "undefined #4") }
         t   = Prod (Map a One) (Prod (Map a c) (Map a d))
mapsum' (Map a (Prod b (Either c d))) 
   = return $ View  -- (Rep (FUN "uncojoin" uncojoin) (FUN "cojoin" (cojoin a))) 
                    (Rep (FUN "uncojoin" (error "undefined #6")) (FUN "cojoin" (error "undefined #7"))) 
                    (TInv (Prod (Map a b) (Prod (Map a c) (Map a d))) (mapsum_inv a b c d))
mapsum' t = mzero

uncojoin :: Map a (b, (Either c d)) -> (Map a b, (Map a c, Map a d))
uncojoin = error "undefined #6"
-- Pre: non duplicate keys
cojoin :: Type a -> (Map a b, (Map a c, Map a d)) -> Map a (b, (Either c d))
cojoin t = error "undefined #7"


-- mapsum_inv :: PF (Pred (Map a b, (Map a c, Map a d)))
mapsum_inv a b c d = AND inv' inv''
  where inv'  = SetInc a (COMP (Map a c) DOM (COMP (Prod (Map a c) (Map a d)) FST SND)) (COMP (Map a b) DOM FST)
        inv'' = SetInc a (COMP (Map a d) DOM (COMP (Prod (Map a c) (Map a d)) SND SND)) (COMP (Map a b) DOM FST)
        -- MISSING INVARIANT: the two domains are mutually exclusive??


-- | Distribute product over sum.
prodsum :: Rule
prodsum (Prod a (Either b c)) 
  = return $ View (Rep (FUN "distr" distr) (FUN "undistr" undistr)) (Either (Prod a b) (Prod a c))
 where
   distr (x, Left y) = Left (x,y)
   distr (x, Right y) = Right (x,y)
   undistr (Left (x,y)) = (x, Left y)
   undistr (Right (x,y)) = (x, Right y)
prodsum (Prod (Either b c) a)
  = return $ View (Rep (FUN "distl" distl) (FUN "undistl" undistl)) (Either (Prod b a) (Prod c a))
 where
          distl (Left y,x) = Left (y,x)
          distl (Right y,x) = Right (y,x)
          undistl (Left (y,x)) = (Left y,x)
          undistl (Right (y,x)) = (Right y,x)
prodsum _ = mzero

summap :: Rule
summap (Map (Either a b) c) 
   = return $ View (Rep (FUN "unpeither" unpeither) (FUN "peither" peither)) (Prod (Map a c) (Map b c))
  where
    unpeither = ( mapKeysMonotonic (either id (error "undefined #7")) . Map.fromDistinctAscList >< 
                  mapKeysMonotonic (either (error "undefined #8") id) . Map.fromDistinctAscList
                ) . List.partition (isLeft . fst) . Map.toAscList
    peither = Map.fromDistinctAscList . uncurry (++) . 
              (fmap (Left >< id) . Map.toAscList >< fmap (Right >< id) . Map.toAscList)
summap _ = mzero





----------------------------------------
-- ** Split, join

-- | Split and join.
mapprodmap :: Rule
mapprodmap (Map a (Prod b (Map c d))) 
  = return $ View mapcomposite (Prod (Map a b) (Map (Prod a c) d))
 where
  -- Precondition to from: all keys a in second table must exist in the first
  -- mapcomposite :: Type a -> Rep (Map a (b, Map c d)) (Map a b, Map (a,c) d)
  mapcomposite = Rep unnjoinProd njoinProd
  unnjoinProd = SPLIT (MAP FST) (COMP (Map a (Map c d)) (FUN "unnjoin" unnjoin) (MAP SND))
  njoinProd = COMP (Prod (Map a b) (Func a (Map c d))) LJOIN (PROD ID (FUN "njoin" njoin))
  unnjoin :: Map a (Map b c) -> Map (a,b) c
  unnjoin = Map.fromDistinctAscList . concat . 
            List.map ((\ (x,l) -> List.map ((const x /\ id) >< id) l) . (id >< Map.toAscList)) . 
            Map.toAscList
  --njoin :: Map (a,b) c -> (a -> (Map b c))
  njoin m x = (Map.fromDistinctAscList . (List.map (snd >< id)) . (List.filter (\ (y,b) -> gcompare a x (fst y) == EQ)) . Map.toAscList) m 
mapprodmap _ = mzero

-- | Split that introduces constraint
split :: Rule
split (Map a (Map c d)) = return $ View rep (TInv t (split_inv a One c d))
  where 
    rep = Rep {  to    = COMP (Map a (Prod One (Map c d))) MapUnJoin2 (MAP (SPLIT BANG ID)), 
                 from  = COMP (Map a (Prod One (Map c d))) (MAP SND) MapJoin2}
    t   = Prod (Map a One) (Map (Prod a c) d)
split (Map a (Prod b (Map c d))) = return $ View rep (TInv t (split_inv a b c d))
  where
    rep = Rep { to = MapUnJoin2, from = MapJoin2 }
    t   = Prod (Map a b) (Map (Prod a c) d)
split t = mzero

split_inv a b c d = 
  SetInc a
    (COMP (Set (Prod a c)) (SET FST) (COMP (Map (Prod a c) d) DOM SND)) 
    (COMP (Map a b) DOM FST)
  
{- 
The case where the map has a swapped product as a key (Map (Prod b a) c), 
is not treated.
-}
-- | Join that matches on invariant
join :: Rule
{-
join (TInv (Prod (Map a One) (Map (Prod _ b) c)) 
     (RInc (RComp (RSetMap RPi1) (RComp RDom RPi2)) (RComp RDom RPi1))) = do
  return $ View rep (Map a (Map b c))
  where rep = Rep { to = RMapFuse1, from = RUnJoin1 }
join (TInv (Prod (Map (Prod a b) c) (Map _ One))
      (RInc (RComp (RSetMap RPi1) (RComp RDom RPi1)) (RComp RDom RPi2))) = do
  return $ View rep (Map a (Map b c))
  where rep = Rep { to = RComp RMapFuse1 RSwap, from = RComp RSwap RUnJoin1}
-}
join (TInv (Prod (Map a b) (Map (Prod _ c) d)) 
           (SetInc _ (COMP _ (SET FST) (COMP _ DOM SND)) (COMP _ DOM FST))) 
  = return $ View rep (Map a (Prod b (Map c d)))
  where rep = Rep { to = MapJoin2, from = MapUnJoin2 }
join (TInv (Prod (Map (Prod a c) d) (Map _ b))
           (SetInc _ (COMP _ (SET FST) (COMP _ DOM FST)) (COMP _ DOM SND))) 
  = return $ View rep (Map a (Prod b (Map c d)))
  where rep = Rep {  to    = COMP (Prod (Map a b) (Map (Prod a c) d)) MapJoin2 (SPLIT SND FST), 
                     from  = COMP (Prod (Map a b) (Map (Prod a c) d)) (SPLIT SND FST) MapUnJoin2}
join t = mzero


{-

----------------------------------------
-- ** Recursion

-- | Type of tables.
type Table f = (Map Int (f Int), Int)

-- | Refine fix-point datatype to table type.
fixastable :: Rule
fixastable (Mu f) = do
  (View (Rep t f) fInt) <- applyF f Int
  let rep = Rep {to = (Map.map t >< id) . fixtotable, from = tabletofix . (Map.map f >< id)}
  return (View rep (Prod (Map Int (fInt)) Int)) 
   where
    fixtotable :: Dist f => Mu f -> Table f
    fixtotable mu = execState (mcata aux mu) (Map.empty,-1)
      where 
        aux :: f Int -> State (Table f) Int
        aux i = do
          (t,k) <- get
          put (Map.insert (k+1) i t, k+1)
          return (k+1)
    -- Precondition to abs: non circular and all keys exist
    tabletofix :: Functor f => Table f -> Mu f
    tabletofix = ana aux
      where 
        aux :: Functor f => Table f -> f (Table f)
        aux (t,k) = fmap (\k -> (t,k)) (fromJust $ Map.lookup k t)
fixastable _ = Nothing

-- | Unlift a functor while applying it to a type.
--   Previously: removefunctors.
applyF :: Fctr f -> Type a -> Maybe (View Rep (Type (f a)))
applyF Id x        = return (View (Rep {to = unIdent, from = Ident}) x)
applyF (K a) x     = return (View (Rep {to = unConst, from = Const}) a)
applyF (g:*:h) x   = do
  (View (Rep g2gx gx2g) gx) <- applyF g x
  (View (Rep h2hx hx2h) hx) <- applyF h x
  let rep = Rep {to = \(Pair x y) -> (g2gx x,h2hx y), from = \(x,y) -> Pair (gx2g x) (hx2h y)}
  return (View rep (Prod gx hx))
applyF (g:+:h) x = do
  (View (Rep g2gx gx2g) gx) <- applyF g x
  (View (Rep h2hx hx2h) hx) <- applyF h x
  let rep = Rep {to = (g2gx-|-h2hx) . aux, from = xua . (gx2g-|-hx2h)}
  return (View rep (Either gx hx))
      where 
        aux (Inl x) = Left x
        aux (Inr x) = Right x
        xua (Left x)  = Inl x
        xua (Right x) = Inr x
applyF (g:@:h) x   = do 
  (View (Rep to from) hx) <- applyF h x
  let rep = Rep (fmap to . unComp) (Comp . fmap from)
  return (View rep (g hx))
-}

-----------------------------------------------------------------------------
-- * Associativity and commutativity

-- | Associate product to the right. One step.
assocr :: Rule
assocr (Prod (Prod a b) c) 
  = return $ View (Rep (assocrPF a b) (assoclPF b c)) (Prod a (Prod b c))
assocr _ = mzero

assocrPF a b = SPLIT (COMP (Prod a b) FST FST) (PROD SND ID) -- fst.fst /\ snd><id
assoclPF b c = SPLIT (PROD ID FST) (COMP (Prod b c) SND SND) -- id><fst /\ snd.snd

-- | Associate product to the left. One step.
assocl :: Rule
assocl (Prod a (Prod b c)) = return $ View (Rep to fro) (Prod (Prod a b) c)
  where
    to  = SPLIT (PROD ID FST) (COMP (Prod b c) SND SND) -- id><fst /\ snd.snd
    fro = SPLIT (COMP (Prod a b) FST FST) (PROD SND ID) -- fst.fst /\ snd><id 
assocl _ = mzero

-- | Put maps to the right and non-maps to the left.
--   Needs to be redesigned....

prods :: Rule
prods = many (many (once assocr) >>> once swap) >>> many (once assocl)
 where
  swap :: Rule
  swap (Prod a (Map b c)) = mzero
  swap (Prod a (Prod (Map b c) d)) = mzero
  swap (Prod (Map a b) c) = return $ View (Rep to fro) (Prod c (Map a b))
    where
      to  = SPLIT SND FST -- snd/\fst
      fro = SPLIT SND FST -- snd/\fst
  swap _ = mzero

-- | Right unit for product.
prodone :: Rule
prodone (Prod a One) = return $ View (Rep to fro) a
 where
    to  = FST              -- fst
    fro = SPLIT ID BANG    -- \x -> (x,())
prodone _ = mzero

-----------------------------------------------------------------------------
-- * Miscelaneous rules

-- | Refine an n-sum of taged ones to an integer.
--   In other words, interpret the n-sum as an enumeration type.
enum2int :: Rule
enum2int = (many (once eitherRightAssoc)) >>> e2i 0 where
   e2i :: Int -> Rule
   e2i i (Tag tg One) = return $ View (Rep t f) Int
      where t   = FUN "e2i" (\() -> i)
            f   = FUN "i2e" (ifeq () (error "Enum invalid") i)
   e2i i (Either (Tag tg One) b) = do
      (View rep b) <- e2i (i+1) b
      Eq <- teq Int b
      let t = EITHER (FUN "e2i" (\() -> i)) (to rep)
      let f = FUN "i2ne" (ifeq (Left ()) (Right . (eval $ from rep)) i)
      return $ View (Rep t f) b
   e2i _ _ = Nothing
   
   ifeq l r i i' = if i==i' then l else r i'
{-
enum2int :: Rule''
enum2int = (many (once eitherRightAssoc)) >>> e2i 0 where
  e2i :: Int -> Rule''
  e2i i (Tag tg One) = do
    return (View (RepPF (PNT i) (FUN "i2eBase" (ifeq () (error "Enum invalid") i))) Int)
  e2i i a@(Either (Tag tg One) b) = do
    (View (RepPF to fro) int) <- e2i (i+1) b
    Eq <- teq Int int
    let to' = VAR ("e2i") -- EITHER (PNT i) (VAR "to") -- to     ---<<< WHY DOESN'T IT WORK???
    let fro' = VAR ("i2e") -- (ifeq (Left ()) (Right . (eval fro)) i)
    return (View (RepPF to' fro') Int)
    -- return (View (RepPF (EITHER (PNT i) to) (FUN ("i2e"++show i) (ifeq (Left ()) (Right . (eval fro)) i))) b)   
  e2i _ _ = Nothing
ifeq l r i i' = if i==i' then l else r i'
-}

-- | Associate sums to the right.
--   Used by 'enum2int' to discover enumeration types.
eitherRightAssoc :: Rule
eitherRightAssoc (Either (Either a b) c) 
   = return $ View (Rep to from) (Either a (Either b c))
  where
      to = EITHER (SUM ID INL) (COMP (Either b c) INR INR)
      from = EITHER (COMP (Either a b) INL INL) (SUM INR ID) 
eitherRightAssoc _ = mzero

-- | Remove all tags from a type represention. This is useful to
--   allow later rules to assume no tags are present.
removetags :: Rule
removetags = many (once aux)
    where aux (Tag n x) = return $ View idrep x
	  aux _ = mzero


-----------------------------------------------------------------------------
-- * Auxiliary types and functions

-- | Determine whether the left alternative is present.
isLeft :: Either a b -> Bool
isLeft = either (const True) (const False)

----------------------------------------
-- ** Pointfree operators

infix 5 ><
infix 4 /\
infix 3 -|-
infix 2 \/

(f >< g) (x,y) = (f x, g y)
(/\) f g x = (f x, g x)
(f -|- g) (Left x) = Left (f x)
(f -|- g) (Right x) = Right (g x)
(\/) f g (Left x) = f x
(\/) f g (Right x) = g x
    

-- new version of rules to transform a RDB into a SS


-- | the invariant representing the FD
--
fdInv t = SetInc (Prod t t) (COMP (List (Prod t t)) LIST2SET CompList)
                            (COMP (List (Prod t t)) LIST2SET ListId)


-- 
table2sstable :: Rule
table2sstable (TInv (Prod (Map a b) (Map c d)) 
                    (SetInc fk (COMP _ pia (COMP _ DOM FST)) (COMP _ pic (COMP _ DOM SND)))) = 
    return $ View rep (TInv (Prod (TInv (List (Prod a b)) (fdInv b))
                                  (TInv (List (Prod c d)) (fdInv d))) inv)
  where
    inv = SetInc fk 
            (COMP (Set a) pia (COMP (List a) LIST2SET (COMP (List (Prod a b)) (LIST FST) FST))) 
            (COMP (Set c) pic (COMP (List c) LIST2SET (COMP (List (Prod c d)) (LIST FST) SND)))
    rep = Rep {to = PROD Table2sstable Table2sstable, from = PROD Sstable2table Sstable2table}            



-----------------------------------------------------------------------------

trySimplify t f = reduce simplifyRel t f
  
trySimplify2 t f = reduce optimize t f 

trySimplify3 t f = reduce simplifyrdb t f 

-- NEED TO TEST THE STRATEGIES


-- strategy to transform a RDB into a SS
--
rdb2ss :: Rule
rdb2ss = simplify >>> (many (aux >>> many aux)) >>> (many (aux2 >>> many aux2)) >>> simplify -- >>> auxxx
  where
    aux = (once (maps2map ||| maps2tables)) >>> simplify

    aux2 = (once map2table) >>> simplify

    simplify = simpleInv
   -- this auxxx is only used in the final function, remove it if not using there 
    auxxx = (many (once (assocr)))


-- just transforms each table (map) into a list 
-- 
rdb2ssDirect :: Rule
rdb2ssDirect = many (once map2table)



-- | simplifying the ss
--
ss2ss :: Rule
ss2ss = rdb2ss >>> ( maps2table) 



-- tries to smplify the invariant
--
simpleInv :: Rule
simpleInv (TInv t i) = 
  do
    View rep tp <- simpleInv t
    let i' = trySimplify3 (Func tp Bool) (COMP t i (from rep))
    return $ View rep  (TInv tp i')

simpleInv (Prod a b) = 
 do
    View repa a' <- simpleInv a
    View repb b' <- simpleInv b
    let froma = from repa
        toa = to repa
        fromb = from repb
        tob = to repb
        to' = trySimplify3 (Func (Prod a b) (Prod a' b')) (PROD toa tob)
        from' = trySimplify3 (Func (Prod a' b') (Prod a b)) (PROD froma fromb)
    return $ View (Rep {to = to', from = from'}) (Prod a' b')

simpleInv t = return $ View idrep t


-- NEED TO TEST ALL THIS RULES

-- rules to transform a a relational database into a spreadsheet 

-- a basic rule: a -> b  <=  [(a, b)] when a functional dependency exists 
-- between a and b in the lis
-- tests seems to be OK
-- testar o inv. da FD com este exemplo: propertyNo -> clientNo x cName
-- propSch1 = Atts $ map Att ["clientNo", "propertyNo", "cName"]
-- propRel1 = [ ["cr76","pg4","john"],
--              ["cr76","pg16","john"]]
--
map2table :: Rule
map2table (TInv (Map a b) i) = 
  do
    let inv' = trySimplify3 (Func (List (Prod a b)) Bool) (AND (COMP (Map a b) i Table2map) inv)

    return $ View rep (TInv (List (Prod a b)) inv')
  where
    rep :: Rep (Map a b) [(a, b)]
    rep = Rep {to = Map2table, from = Table2map}  
    -- img (y . R . x) includedIn id
    inv = SetInc (Prod b b) (COMP (List (Prod b b)) LIST2SET CompList) 
                            (COMP (List (Prod b b)) LIST2SET ListId)
      -- ListInc (Prod b b) CompList ListId
 
map2table (Map a b) = 
  do
    return $ View rep (TInv (List (Prod a b)) inv)
  where
    rep :: Rep (Map a b) [(a, b)]
    rep = Rep {to = Map2table, from = Table2map}
    -- img (y . R . x) includedIn id
    inv = SetInc (Prod b b) (COMP (List (Prod b b)) LIST2SET CompList) 
                            (COMP (List (Prod b b)) LIST2SET ListId)
    --ListInc (Prod b b) CompList ListId

map2table _ = mzero



-- | ([A * B], [C * D]) ---> [A * B * C * D]
--   with the other rules I get the first pair but I need to get to the most simple model
--   only the I can go from the ss to the rdb
--
maps2table :: Rule

maps2table (Prod (List (Prod a b)) (List (Prod c d))) =
  (do
    Eq <- teq a c
    Eq <- teq b d
    return $ View rep (List (Prod a b)))
  `mplus`
  (do
    Eq <- teq a d
    Eq <- teq b c
    return $ View rep (List (Prod a b)))
  `mplus`
  (do
    Eq <- teq a c
    return $ View rep (List (Prod a (Prod b d))))
  `mplus`
  (do
    Eq <- teq a d
    return $ View rep (List (Prod a (Prod b c))))
  `mplus`
  (do
    Eq <- teq b c
    return $ View rep (List (Prod a (Prod b d))))
  `mplus`
  (do
    Eq <- teq b d
    return $ View rep (List (Prod a (Prod b c))))
  `mplus` 
  (do
    return $ View rep (List (Prod a (Prod b (Prod c d)))))
  where
    rep = undefined

maps2table (TInv t i) = (maps2table) t

maps2table (Prod (TInv a i) b) = maps2table (Prod a b)

maps2table (Prod a (TInv b i)) = maps2table (Prod a b)

maps2table t = return $ View idrep t

{-
maps2table (Prod (List (Prod a b)) (List (Prod c d))) =
  do
    return $ View rep (List (Prod a (Prod b (Prod c d))))
where
    rep :: Rep ([(a, b)], [(c, d)]) [(a, (b, (c, d)))]
    rep = Rep {to = Maps2table, from = Table2maps}
-}

maps2table _ = mzero    




-- (A -> B) * (A -> C)  <=  (A -> (B * C?))
-- there is a foreign key between A in the second map and A in the first one
-- tests seems to be ok
--
maps2map :: Rule

maps2map (TInv (Prod (Map a b) (Map _ c))
               (SetInc _ (COMP _ DOM SND) (COMP _ DOM FST))) = 
  do 
    return $ View rep (Map a (Prod b (Maybe c)))
  where
    rep :: Rep (Map a b, Map a c) (Map a (b, Maybe c))
    rep = Rep {to = Maps2map, from = Map2maps}

-- same as above but with the pair swapped
maps2map (TInv (Prod (Map a c) (Map _ b))
               (SetInc _ (COMP _ DOM FST) (COMP _ DOM SND))) = 
  do 
    return $ View rep (Map a (Prod c (Maybe b)))
  where
    rep :: Rep (Map a c, Map a b) (Map a (c, Maybe b))
    rep = Rep {to = Maps2map, from = Map2maps}


-- just PART of the key is refered
-- c ---> a
maps2map (TInv (Prod (Map a b) (Map c d))
               (SetInc _ (COMP _ DOM SND) (COMP _ _ (COMP _ DOM FST)))) = 
  do
    return $ View rep (Map a (Prod b (Maybe d)))
  where
    rep :: Rep (Map a b, Map c d) (Map a (b, Maybe d))
    rep = Rep {to = Maps2map, from = Map2maps}

-- a ---> c
-- same as above but with the foreign key in the other direction
maps2map (TInv (Prod (Map a b) (Map c d))
               (SetInc _ (COMP _ DOM FST) (COMP _ _ (COMP _ DOM SND)))) = 
  do
    return $ View rep (Map c (Prod d (Maybe b)))
  where
    rep :: Rep (Map a b, Map c d) (Map c (d, Maybe b))
    rep = Rep {to = Maps2map', from = Map2maps'}

maps2map _ = mzero

{-
maps2map (TInv (Prod (Map a b) (Map _ c))
               (SetInc _ (COMP _ DOM SND) (COMP _ DOM FST))) = 
  do 
    return $ View rep (Map a (Prod b (Maybe c)))
  where
    rep :: Rep (Map a b, Map a c) (Map a (b, Maybe c))
    rep = Rep {to = Maps2map, from = Map2maps}

-- same as above but with the pair swapped
maps2map (TInv (Prod (Map a c) (Map _ b))
               (SetInc _ (COMP _ DOM FST) (COMP _ DOM SND))) = 
  do 
    return $ View rep (Map a (Prod c (Maybe b)))
  where
    rep :: Rep (Map a c, Map a b) (Map a (c, Maybe b))
    rep = Rep {to = Maps2map, from = Map2maps}


-- ((A * D) -> B) * (A -> C)  <=  (A -> (B * C?))
---  ^---------------|
-- the key is not all refered, just part of it
maps2map (TInv (Prod (Map (Prod a b) c) (Map _ d))
               (SetInc _ (COMP _ DOM SND) (COMP _ (SET FST) (COMP _ DOM FST)))) = 
  do
    return $ View rep (Map (Prod a b) (Prod c (Maybe d)))
  where
    rep = Rep {to = Maps2map', from = Map2maps'}


-- same as above but with the pair swapped
maps2map (TInv (Prod (Map _ d) (Map (Prod a b) c))
               (SetInc _ (COMP _ DOM FST) (COMP _ (SET FST) (COMP _ DOM SND)))) = 
  do
    return $ View rep (Map (Prod a b) (Prod c (Maybe d)))
  where
    rep = Rep {to = Maps2map'', from = Map2maps''}
-}



-- (A -> B) * ((A * C) -> D)  <=  [(A * B)] * [(A * C * D)]
--  ^           |                   ^           |
--  |-----------|                   |___________|
-- tests seem to OK
--
maps2tables :: Rule
-- c ---> a
maps2tables (TInv (Prod (Map a b) (Map c d))
                  (SetInc t (COMP _ x (COMP _ DOM SND)) (COMP _ DOM FST))) = 
    do 
    return $ View rep (TInv (Prod (List (Prod a b)) (List (Prod c d))) inv)
  where
    fres = List (Prod a b)
    sres = List (Prod c d)

    inv = SetInc t (COMP (Set c) x (COMP (List c) LIST2SET (COMP sres (LIST FST) SND))) 
                   (COMP (List a) LIST2SET (COMP fres (LIST FST) FST))

    rep :: Rep (Map a b, Map c d) ([(a, b)], [(c, d)])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}

-- a ---> c
-- same as above but with the foreign key in the other direction
maps2tables (TInv (Prod (Map a b) (Map c d))
                  (SetInc t (COMP _ x (COMP _ DOM FST)) (COMP _ DOM SND))) = 
    do 
    return $ View rep (TInv (Prod (List (Prod a b)) (List (Prod c d))) inv)
  where
    fres = List (Prod a b)
    sres = List (Prod c d)

    inv = SetInc t (COMP (Set a) x (COMP (List a) LIST2SET (COMP fres (LIST FST) FST))) 
                   (COMP (List c) LIST2SET (COMP sres (LIST FST) SND))

    rep :: Rep (Map a b, Map c d) ([(a, b)], [(c, d)])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}

-- PART of one key reperences PART of another key
-- c ---> a
maps2tables (TInv (Prod (Map a b) (Map c d))
                  (SetInc t (COMP _ x (COMP _ DOM SND))
                            (COMP _ y (COMP _ DOM FST)))) = 
    do 
      return $ View rep (TInv (Prod (List (Prod a b)) (List (Prod c d))) inv)
  where
    fres = List (Prod a b)
    sres = List (Prod c d)

    inv = SetInc t (COMP (Set c) x (COMP (List c) LIST2SET (COMP sres (LIST FST) SND)))
                   (COMP (Set a) y (COMP (List a) LIST2SET (COMP fres (LIST FST) FST)))
               
    rep :: Rep (Map a b, Map c d) ([(a, b)], [(c, d)])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}

-- a ---> c
-- same as above but with the foreign key in the other direction
maps2tables (TInv (Prod (Map a b) (Map c d))
                  (SetInc t (COMP _ x (COMP _ DOM FST))
                            (COMP _ y (COMP _ DOM SND)))) = 
    do 
      return $ View rep (TInv (Prod (List (Prod a b)) (List (Prod c d))) inv)
  where
    fres = List (Prod a b)
    sres = List (Prod c d)

    inv = SetInc t (COMP (Set a) x (COMP (List a) LIST2SET (COMP fres (LIST FST) FST)))
                   (COMP (Set c) y (COMP (List c) LIST2SET (COMP sres (LIST FST) SND)))
               
    rep :: Rep (Map a b, Map c d) ([(a, b)], [(c, d)])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}


-- | A -> B  *  B -> C   <=   [A * B] * [B * C]
--        |-----^
--   the column B of the first table is a FK to th key of the second table

maps2tables (TInv (Prod (Map a b) (Map _ c)) (SetInc t (COMP _ RNG FST) (COMP _ DOM SND))) = 
  do
    return $ View rep (TInv (Prod (TInv (List (Prod a b)) (fdInv b)) (TInv (List (Prod b c)) (fdInv c))) inv)
  where
    inv = SetInc b (COMP (List b) LIST2SET (COMP (List (Prod a b)) (LIST SND) FST)) 
                   (COMP (List b) LIST2SET (COMP (List (Prod b c)) (LIST FST) SND))

    fdInv fdt =  SetInc (Prod fdt fdt) (COMP (List (Prod fdt fdt)) LIST2SET CompList) 
                                   (COMP (List (Prod fdt fdt)) LIST2SET ListId)

    rep :: Rep (Map a b, Map b c) ([(a, b)], [(b, c)])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}

maps2tables (TInv (Prod (Map b c) (Map a b')) (SetInc t (COMP _ RNG SND) (COMP _ DOM FST))) = 
  do
    return $ View rep (TInv (Prod (TInv (List (Prod b c)) (fdInv c)) (TInv (List (Prod a b)) (fdInv b'))) inv)
  where
    inv = SetInc b (COMP (List b) LIST2SET (COMP (List (Prod a b)) (LIST SND) SND)) 
                   (COMP (List b) LIST2SET (COMP (List (Prod b c)) (LIST FST) FST))

    fdInv b = SetInc (Prod b b) (COMP (List (Prod b b)) LIST2SET CompList) 
                                (COMP (List (Prod b b)) LIST2SET ListId)                   

    rep :: Rep (Map b c, Map a b) ([(b, c)], [(a, b)])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}

maps2tables (TInv (Prod (Map b c) (Map a b'))
                  (SetInc _ (COMP t x (COMP _ RNG SND)) (COMP _ DOM FST))) =
  do 
    return $ View rep (TInv (Prod (TInv (List (Prod b c)) (fdInv c)) (TInv (List (Prod a b')) (fdInv b'))) inv)
  where
    inv = SetInc b (COMP t x (COMP (List b') LIST2SET (COMP (List (Prod a b')) (LIST SND) SND)))
                   (COMP (List b) LIST2SET (COMP (List (Prod b c)) (LIST FST) FST))
    
    fdInv b = SetInc (Prod b b) (COMP (List (Prod b b)) LIST2SET CompList) 
                                (COMP (List (Prod b b)) LIST2SET ListId)

    rep :: Rep (Map b c, Map a b') ([(b, c)], [(a, b')])
    rep = Rep {to = PROD Map2table Map2table, from = PROD Table2map Table2map}

maps2tables _ = mzero




{-
maps2tables (TInv (Prod (Map a b) (Map (Prod _ c) d))
                  (SetInc _ (COMP _ (SET FST) (COMP _ DOM SND)) (COMP _ DOM FST))) = 
    do 
    return $ View rep (TInv res inv)
  where
    fres = List (Prod a b)
    sres = List (Prod (Prod a c) d)
    res = Prod fres sres

    inv = SetInc a (COMP (List a) LIST2SET (COMP sres (LIST (COMP (Prod a c) FST FST)) SND)) 
                   (COMP (List a) LIST2SET (COMP fres (LIST FST) FST))
    --ListInc a (COMP sres (LIST (COMP (Prod a c) FST FST)) SND) (COMP fres (LIST FST) FST)

    rep :: Rep (Map a b, Map (a, c) d) ([(a, b)], [((a, c), d)])
    rep = Rep {to = Maps2tables, from = Tables2maps}

-- same as above but with the pair swapped
maps2tables (TInv (Prod (Map (Prod _ c) d) (Map a b))
                  (SetInc _ (COMP _ (SET FST) (COMP _ DOM FST)) (COMP _ DOM SND))) = 
    do 
      return $ View rep (TInv res inv)
  where
    fres = List (Prod (Prod a c) d)
    sres = List (Prod a b)
    res = Prod fres sres

    inv = SetInc a (COMP (List a) LIST2SET (COMP fres (LIST (COMP (Prod a c) FST FST)) FST))
                   (COMP (List a) LIST2SET (COMP sres (LIST FST) SND))
    --ListInc a (COMP fres (LIST (COMP (Prod a c) FST FST)) FST) (COMP sres (LIST FST) SND)
               
    rep :: Rep (Map (a, c) d, Map a b) ([((a, c), d)], [(a, b)])
    rep = Rep {to = Maps2tables', from = Tables2maps'}
    
    
-- PART of one key reperences PART of another key
maps2tables (TInv (Prod (Map (Prod a b) c) (Map (Prod _ d) e))
                  (SetInc _ (COMP _ (COMP _ (SET FST) DOM) SND) 
                            (COMP _ (COMP _ (SET FST) DOM) FST))) = 
    do 
      return $ View rep (TInv res inv)
  where
    fres = List (Prod (Prod a b) c)
    sres = List (Prod (Prod a d) e)
    res = Prod fres sres

    inv = SetInc a (COMP (List a) LIST2SET (COMP sres (LIST (COMP (Prod a d) FST FST)) SND))
                   (COMP (List a) LIST2SET (COMP fres (LIST (COMP (Prod a b) FST FST)) FST))
    --ListInc a (COMP sres (LIST (COMP (Prod a d) FST FST)) SND) 
    --                (COMP fres (LIST (COMP (Prod a b) FST FST)) FST)
               
    rep :: Rep (Map (a, b) c, Map (a, d) e) ([((a, b), c)], [((a, d), e)])
    rep = Rep {to = Maps2tables'', from = Tables2maps''}

-- same as above but with the pair swapped
maps2tables (TInv (Prod (Map (Prod _ d) e) (Map (Prod a b) c))
                  (SetInc _ (COMP _ (SET FST) (COMP _ DOM FST)) 
                            (COMP _ (SET FST) (COMP _ DOM SND)))) = 
    do 
      return $ View rep (TInv res inv)
  where
    fres = List (Prod (Prod a d) e)
    sres = List (Prod (Prod a b) c)
    res = Prod fres sres

    inv = SetInc a (COMP (List a) LIST2SET (COMP fres (LIST (COMP (Prod a d) FST FST)) FST))
                   (COMP (List a) LIST2SET (COMP sres (LIST (COMP (Prod a b) FST FST)) SND))
    --ListInc a (COMP fres (LIST (COMP (Prod a d) FST FST)) FST) 
      --              (COMP sres (LIST (COMP (Prod a b) FST FST)) SND)
               
    rep :: Rep (Map (a, d) e, Map (a, b) c) ([((a, d), e)], [((a, b), c)])
    rep = Rep {to = Maps2tables'', from = Tables2maps''}
-}


