module HaExcel.HaExcel( 
         {-      module Sheet.Expr,
               module UCheck.MParser,
               module Sheet.Sheet,
               module Sheet.Indx,
               module UCheck.UnitErrors,
               module UCheck.Units,
               module UCheck.FaultLoc,
               module UCheck.Header, -}
               module Tests,
               module FDInference,
               module FDCare,
               module Types,
               module Formulas,
               module Synthesize,
               module HaExcel.FDS2Type,
               module HaExcel.ReadSS,
               module HaExcel.GenSS,
               module HaExcel.GenSQL,
               module HaExcel.GenDB,
               module HaExcel.HaExcel,
               ss2boxes,
               ss2rdb,
               ss2SQL,
               ss2DB,
               ss2SS
             ) where


import Control.Monad
import Debug.Trace
import Data.List
import Data.Maybe as Maybe
import System.IO

import FDCare
import FDInference hiding (transProjection)
import Synthesize(synthesize)
import Types 
import Tests
import Formulas

import HaExcel.FDS2Type 
import HaExcel.ReadSS
import HaExcel.GenSQL
import HaExcel.GenDB hiding (stripSpace, showT, stripForm)
import HaExcel.GenSS

import Data.Transform.TwoLevelPF hiding (teq, Func)
import Data.Transform.PropExample

import Sheet.Expr hiding (Eq)
import Sheet.Sheet
import Sheet.Indx
import UCheck.MParser
import UCheck.Header hiding (Table)
import UCheck.UnitErrors
import UCheck.Units
import UCheck.FaultLoc (faultLoc)
import Language.Gnumeric.Syntax

import Database.HaskellDB.Sql
import Database.HaskellDB.HSQL.MySQL
import Text.XML.HaXml.Xml2Haskell


-- | export to SQL as strings
--
ss2SQL :: FilePath -> FilePath -> IO ()
ss2SQL f fout = do 
                  x <- ss2boxes f 
                  let y = genSQL x
                  o <- openFile fout WriteMode
                  mapM (mapM (hPutStrLn o)) y
                  hFlush o
                  hClose o


-- | returns an HaskellDB MySQL DB
--
ss2DB :: FilePath -> MySQLOptions -> IO ()
ss2DB f m = ss2boxes f >>= \x -> genDB x m


-- | return box with RDB type inside, forgots the Sheet part
--
ss2rdb :: FilePath -> IO [Maybe Box]
ss2rdb f = ss2boxes f >>= return . map fst


-- | exporting to Gnumeric
--    given a filepath to a spreadsheet and a filepath to store the new SS, it will
--    tranform the SS into a RDB, goes back to the SS and migrates the data.
--    the resulting SS has the same aspect as the RDB
--
ss2SS :: FilePath -> FilePath -> IO ()
ss2SS f fout = do
  -- read the gnumeric file and get a list of annoted Sheets 
          l <- readSS f
          (wb ::  Gmr'Workbook) <- readGnumeric' f
  -- just select the "DB" ones          
          let (_, mmet) = break (\sh -> ((howManyInts sh) * 2) <= (length $ unSheet sh)) l
  -- calculate the expressions and thransform it into lists of list of strings          
              ucf = map (sheet2CompleteList . calcExpr) mmet
              ucf' = map (sheet2CompleteList . sheet2trigger) mmet
  -- separate the attributes from the data              
              mmet' = map (\l -> (Atts $ map (Att ) $ head l, tail l)) ucf
              mmet'' = map (\l -> (Atts $ map (Att ) $ head l, tail l)) ucf'
  -- find the FDs               
              fdstype = map (uncurry rel2rdb) mmet'
          --    fd = fun propSch2  propRel2 >>= lks2fds 
          --    fdstype = [fds2type $ rel2rdb1 (fd ++ fdds1) propSch2 propRel2]
  -- transform the FDs into a Type              
             -- fdstype = map fds2type fds
              boxs = map (transformData' rdb2ss) (zip fdstype mmet')
              wb' = genSS boxs wb
        --  mapM (putStrLn . show) boxs                          
          o <- openFile fout WriteMode
          hPutXml o wb'
          hFlush o
          hClose o


-- | given a rule to transform the RDB into a SS and a relation, it will compute the SS
--    and return a box with the data in the SS format
--
transformData' :: Rule -> (Types.Box, (R String, Relation String)) -> Maybe Types.Box
transformData' rul (Box rdb _, (r, rl)) = 
  do
    View rep ss <- rul rdb
    getData (Box ss undefined) r rl



-- | returns a list with pair of box with the calculated SS and the SS with SQL triggers
--
ss2boxes :: FilePath -> IO [(Maybe Box, Maybe Box)] -- [[String]]
ss2boxes f = do
  -- read the gnumeric file and get a list of annoted Sheets 
          l <- readSS f
  -- just select the "DB" ones          
          let (_, mmet1) = break (\sh -> ((howManyInts sh) * 2) <= (length $ unSheet sh)) l
              mmet = [head mmet1]
  -- calculate the expressions and thransform it into lists of list of strings          
              ucf = map (sheet2CompleteList . calcExpr) mmet
              ucf' = map (sheet2CompleteList . sheet2trigger) mmet
  -- separate the attributes from the data              
              mmet' = map (\l -> (Atts $ map (Att . stripSpace) $ head l, tail l)) ucf
              mmet'' = map (\l -> (Atts $ map (Att . stripSpace) $ head l, tail l)) ucf'
  -- find the FDs               
              fdstype = map (uncurry rel2rdb) mmet'
             -- fd = fun propSch2  propRel2 >>= lks2fds 
             -- fdstype = [fds2type $ rel2rdb1 (fd ++ fdds1) propSch2 propRel2]

  -- transform the FDs into a Type              
           --    fdstype = map fds2type fds
  -- transform the data according to a rule              
              dat = map (transformData rdb2ss) (zip fdstype [(propSch2, propRel2)])
              dat' = map (transformData rdb2ss) (zip fdstype mmet'')
  -- zip each box with the corresponding Sheet and remove the ones that failed to transform
             -- zp = filter ((maybe False (const True)) . fst) $ zip dat dat'
              zp = zip dat dat'
  -- remove the Just
              -- sql = map (\(Just b, Just s) -> (b, s)) zp 
          return zp -- $ showSQL sql                



-- |  receives a box with the RDB and returns the SS accordign to a rule
--
transformData :: Rule -> (Types.Box, (R String, Relation String)) -> Maybe Types.Box
transformData rul (Box t _, (r, rl)) = 
  do
    View rep t' <- rul t
    (Box td d) <- getData (Box t' undefined) r rl
    Eq <- teq td t'
    return $ Box t ((eval $ from rep) d)


-- | aux stuff
--

stripSpace = map (\c -> if c == ' ' then '_' else c)


-- | removes the FORM prefix from an att
--
stripForm s = maybe s id (stripPrefix "FORM" s)


pfile f = "HaExcel/Examples/Gnumeric/" ++ f ++ ".gnumeric"



xp = do 
      (r, rl, _) <- readGnu (pfile "prop2") >>= return . head
      let fs = fun r rl >>= lks2fds
          fdds = [(Atts $ map Att ["FORMtotalDays", "rentPerDay"], Atts $ map Att ["FORMtotal rent"]),
                  (Atts $ map Att ["rentStart", "rentFinish"], Atts $ map Att ["FORMtotalDays"])]
      showfds $ rel2rdb1 (fs ++ fdds1) r rl
--      return r


x1 = Atts $ map Att ["pNo", "cNo", "pNa", "cNa"] 

x1d = [["P1", "C1", "A", "C"],
      ["P2", "C1", "B", "C"],
      ["P1", "C2", "A", "D"],
      ["P2", "C2", "B", "D"],
    ["P3", "C1", "B", "C"]]


x2 = Atts $ map Att ["pNo", "cNo", "pNa", "cNa", "X"] 

x2d = [["P1", "C1", "A", "C", "1"], 
       ["P2", "C1", "B", "C", "2"], 
       ["P1", "C2", "A", "D", "3"], 
       ["P2", "C2", "B", "D", "4"], 
       ["P3", "C1", "B", "C", "5"]]    

x2d1 = [["P1", "C1", "A", "C", "1"], ["P2", "C1", "B", "C", "2"],  ["P1", "C2", "A", "D", "3"], ["P2", "C2", "B", "D", "4"], ["P3", "C1", "B", "C", "5"], ["P1", "C1", "A", "C", "1"], ["P2", "C1", "B", "C", "2"],  ["P1", "C2", "A", "D", "3"], ["P2", "C2", "B", "D", "4"],  ["P3", "C1", "B", "C", "5"]]       


x2d2 = [["P1", "C1", "A", "C", "1"], 
        ["P2", "C1", "B", "C", "2"],  
        ["P1", "C2", "A", "D", "3"], 
        ["P2", "C2", "B", "D", "4"], 
        ["P3", "C1", "B", "C", "4"], 
        ["P1", "C1", "A", "C", "1"], 
        ["P2", "C1", "B", "C", "2"],  
        ["P1", "C2", "A", "D", "3"], 
        ["P2", "C2", "B", "D", "4"],  
        ["P3", "C1", "B", "C", "4"]]

fdds1 = [(Atts $ map Att ["rentStart", "rentFinish", "rentPerDay"], Atts $ map Att ["FORMtotal rent", "FORMtotalDays"])]


bh = [(Atts $ map Att ["clientNo", "propertyNo"], Atts $ map Att ["rentStart", "rentFinish", "FORMtotal rent", "FORMtotalDays"])]


